@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -46,6 +46,7 @@ EXTERN char *dosemu_tmpdir_process_path INIT("~/" LOCALDIR_BASE_NAME "/tmp/<pid>
 EXTERN char *dosemu_rundir_path INIT("~/" LOCALDIR_BASE_NAME "/run");
 EXTERN char *dosemu_localdir_path INIT("~/" LOCALDIR_BASE_NAME);
 EXTERN char *dosemu_lib_dir_path INIT(DOSEMULIB_DEFAULT);
+EXTERN char *dosemu_hdimage_dir_path INIT(DOSEMULIB_DEFAULT);
 EXTERN char *keymap_load_base_path INIT(DOSEMULIB_DEFAULT "/");
 EXTERN char *keymap_dir_path INIT("keymap/");
 EXTERN char *owner_tty_locks INIT("uucp");
@@ -66,6 +67,7 @@ EXTERN char *dosemu_midi_path INIT("~/" LOCALDIR_BASE_NAME "/run/" DOSEMU_MIDI);
 #define    RUNDIR                dosemu_rundir_path
 #define    LOCALDIR              dosemu_localdir_path
 #define    DOSEMU_LIB_DIR        dosemu_lib_dir_path
+#define    DOSEMU_HDIMAGE_DIR    dosemu_hdimage_dir_path
 #define    KEYMAP_LOAD_BASE_PATH keymap_load_base_path
 #define    KEYMAP_DIR            keymap_dir_path
 #define    OWNER_LOCKS           owner_tty_locks
@@ -126,14 +128,6 @@ EXTERN char *dosemu_midi_path INIT("~/" LOCALDIR_BASE_NAME "/run/" DOSEMU_MIDI);
  */
 #undef __LINE__
 
-/* DEFINE this, if you want the extended vm86 support (vm86plus)
- * You then either have to patch your kernel or load emumodule
- */
-#undef REQUIRES_VM86PLUS
-
-/* DEFINE this, if you have vm86plus built into the kernel */
-#undef BUILTIN_VM86PLUS
-
 /* DPMI test windows */
 #undef WANT_WINDOWS
 
@@ -145,6 +139,9 @@ EXTERN char *dosemu_midi_path INIT("~/" LOCALDIR_BASE_NAME "/run/" DOSEMU_MIDI);
 #undef X2_SUPPORT
 #undef X_GRAPHICS
 
+/* Define this to use the old Slang keyboard code */
+#undef USE_OLD_SLANG_KBD
+
 /*
  *  DEFINE this, if you have joystick support in the kernel
  *  or have the joystick module loaded (needing <linux/joystick.h>)
@@ -165,9 +162,6 @@ EXTERN char *dosemu_midi_path INIT("~/" LOCALDIR_BASE_NAME "/run/" DOSEMU_MIDI);
 /* Define this if you want to use MITSHM */
 #undef HAVE_MITSHM
 
-/* Define this if you want to use a monoton micro timing (Bernd Paysan) */
-#undef MONOTON_MICRO_TIMING
-
 /* Define this if you want dosemu permanently run as (suid) root and only gain
  * user privileges, when accessing secure relevant resources.
  * Otherwise dosemu will permanently run as user and only temporary take over
