@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 #include <stdio.h>
 #include <termios.h>
 #include <unistd.h>
@@ -12,27 +18,28 @@
 
 #include "config.h"
 #include "emu.h"
+#include "bios.h"
 #include "mouse.h"
 #include "video.h"
 #include "timers.h"
 #include "int.h"
 #include "shared.h"
-#include "../dpmi/dpmi.h"
+#include "dpmi.h"
 #include "pic.h"
 #include "ipx.h"
 #include "pktdrvr.h"
+#include "iodev.h"
 
-#ifdef NEW_KBD_CODE
 #include "keyb_clients.h"
+
+#ifdef USE_SBEMU
+#include "sound.h"
 #endif
 
 extern void keyb_server_run(void);
 extern void irq_select(void);
 extern int type_in_pre_strokes();
 
-#ifdef __NetBSD__
-extern int errno;
-#endif
 
 /* Variables for keeping track of signals */
 #define MAX_SIG_QUEUE_SIZE 50
@@ -48,7 +55,7 @@ static struct SIGNAL_queue signal_queue[MAX_SIG_QUEUE_SIZE];
 static sigset_t oldset;
 
 #ifdef __linux__
-#if __GLIBC__ > 1		/* glibc-2 version */
+#if GLIBC_VERSION_CODE >= 2000
 /*
  * Thomas Winder <thomas.winder@sea.ericsson.se> wrote:
  * glibc-2 uses a different struct sigaction type than the one used in
@@ -64,7 +71,7 @@ int
 dosemu_sigaction(int sig, struct sigaction *new, struct sigaction *old)
 {
   struct my_sigaction {
-    __sighandler_t sa_handler;
+    __sighandler_t my_sa_handler;
     unsigned long sa_mask;
     unsigned long sa_flags;
     void (*sa_restorer)(void);
@@ -72,7 +79,7 @@ dosemu_sigaction(int sig, struct sigaction *new, struct sigaction *old)
 
   struct my_sigaction my_sa;
 
-  my_sa.sa_handler = new->sa_handler;
+  my_sa.my_sa_handler = new->sa_handler;
   my_sa.sa_mask = *((unsigned long *) &(new->sa_mask));
   my_sa.sa_flags = new->sa_flags;
   my_sa.sa_restorer = new->sa_restorer;
@@ -112,23 +119,6 @@ signal_init(void)
 {
   struct sigaction sa;
   sigset_t trashset;
-#ifdef __NetBSD__
-  struct sigaltstack salt;
-
-  /* Point to the top of the stack, minus 4
-     just in case, and make it aligned  */ 
-  salt.ss_sp = (char *)cstack;
-  salt.ss_size = sizeof(cstack);
-  salt.ss_flags = 0;
-  if (sigaltstack(&salt, 0) != 0) {
-      /* Aieee! */
-      fprintf(stderr, "cannot set signal handling stack: %s\n\r",
-	      strerror(errno));
-      fflush(stdout);
-      fflush(stderr);
-      _exit(1);
-  }
-#endif
 
   /* block no additional signals (i.e. get the current signal mask) */
   sigemptyset(&trashset);
@@ -136,8 +126,40 @@ signal_init(void)
   g_printf("Initialized all signals to NOT-BLOCK\n");
 
 
-  /* init signal handlers */
-
+  /* init signal handlers - these are the defined signals:
+   ---------------------------------------------
+   SIGHUP		 1	S	leavedos
+   SIGINT		 2	S	leavedos
+   SIGQUIT		 3	S	sigquit
+   SIGILL		 4	N	dosemu_fault
+   SIGTRAP		 5	N	dosemu_fault
+   SIGABRT		 6	S	leavedos
+   SIGBUS		 7	N	dosemu_fault
+   SIGFPE		 8	N	dosemu_fault
+   SIGKILL		 9	na
+   SIGUSR1		10	?	threads
+   SIGSEGV		11	N	dosemu_fault
+   SIGUSR2		12	NQ	(SIG_ACQUIRE)
+   SIGPIPE		13
+   SIGALRM		14	NQ	(SIG_TIME)sigalrm
+   SIGTERM		15	S	leavedos
+   SIGSTKFLT		16
+   SIGCHLD		17	unused?
+   SIGCONT		18
+   SIGSTOP		19
+   SIGTSTP		20
+   SIGTTIN		21	unused	was SIG_SER??
+   SIGTTOU		22
+   SIGURG		23
+   SIGXCPU		24
+   SIGXFSZ		25
+   SIGVTALRM		26
+   SIGPROF		27	N	(SIG_CALIB)sigstretch
+   SIGWINCH		28	NQ	(SIG_RELEASE)
+   SIGIO		29	NQ	sigio
+   SIGPWR		30
+   SIGUNUSED		31	na
+  ------------------------------------------------ */
   NEWSETSIG(SIGILL, dosemu_fault);
   NEWSETQSIG(SIG_TIME, sigalrm);
   NEWSETSIG(SIGFPE, dosemu_fault);
@@ -147,9 +169,6 @@ signal_init(void)
   NEWSETSIG(SIGBUS, dosemu_fault);
 #endif
   SETSIG(SIGINT, leavedos);   /* for "graceful" shutdown for ^C too*/
-#ifdef __NetBSD__
-  NEWSETSIG(SIGURG, vm86_return);
-#endif
   SETSIG(SIGHUP, leavedos);	/* for "graceful" shutdown */
   SETSIG(SIGTERM, leavedos);
 #if 0 /* Richard Stevens says it can't be caught. It's returning an
@@ -158,7 +177,17 @@ signal_init(void)
   SETSIG(SIGKILL, leavedos);
 #endif
   SETSIG(SIGQUIT, sigquit);
-  SETSIG(SIGWINCH, gettermcap); /* Adjust window sizes in DOS */
+
+#ifdef X_SUPPORT
+  if(config.X) {
+    SETSIG(SIGWINCH, SIG_IGN);
+  }
+  else
+#endif
+  {
+    SETSIG(SIGWINCH, gettermcap); /* Adjust window sizes in DOS */
+  }
+
 /*
   SETSIG(SIGUNUSED, timint);
 */
@@ -217,6 +246,7 @@ sti(void)
  */
 void handle_signals(void) {
   if ( SIGNAL_head != SIGNAL_tail ) {
+    signal_pending = 0;
     signal_queue[SIGNAL_head].signal_handler();
     SIGNAL_head = (SIGNAL_head + 1) % MAX_SIG_QUEUE_SIZE;
 /* 
@@ -224,6 +254,7 @@ void handle_signals(void) {
  * by the kernel ASAP.
  */
       if (SIGNAL_head != SIGNAL_tail) {
+        signal_pending = 1;
 	if (in_dpmi)
 	  dpmi_eflags |= VIP;
         REG(eflags) |= VIP;
@@ -252,38 +283,43 @@ void handle_signals(void) {
  * counter here.
  * ============================================================== */
 
-void SIGALRM_call(void){
-
+void SIGALRM_call(void)
+{
+  static int first = 0;
+  static hitimer_t cnt200 = 0;
+  static hitimer_t cnt1000 = 0;
   static volatile int running = 0;
-  static int partials = 0;
 #if VIDEO_CHECK_DIRTY
   static int update_pending = 0;
 #endif
   int retval;
   
-#if defined(SIG) && defined(REQUIRES_VM86PLUS)
+  if (first==0) {
+    cnt200 =
+    cnt1000 =
+    pic_sys_time;	/* initialize */
+    first = 1;
+  }
+
+#if defined(SIG)
   irq_select();  /* we need this in order to catch lost IRQ-SIGIOs */
 #endif
 
 #ifdef X_SUPPORT
   if (config.X) {
      X_handle_events();
-#ifdef NEW_KBD_CODE
      /* although actually the event handler handles the keyboard in X, keyb_client_run
       * still needs to be called in order to handle pasting.
       */
      keyb_client_run();
-#endif
   }
 #endif
 
-#ifdef NEW_KBD_CODE
   /* for other front-ends, keyb_client_run() is called from ioctl.c if data is
    * available, so we don't need to do it here.
    */
   
   keyb_server_run();
-#endif
    
 
 
@@ -362,6 +398,11 @@ void SIGALRM_call(void){
   if (mice->intdrv)
     mouse_curtick();
 
+#ifdef USE_SBEMU
+  /* This is a macro */
+  run_sb();
+#endif
+
   /* TRB - perform processing for the IPX Asynchronous Event Service */
 #ifdef IPX
   if (config.ipxsup)
@@ -370,6 +411,7 @@ void SIGALRM_call(void){
 
   timer_tick();
 
+#if 0
 /*
  * DANG_BEGIN_REMARK
  *  Check for keyboard coming from client
@@ -381,36 +423,46 @@ void SIGALRM_call(void){
    pic_request (PIC_IRQ1);
    *(u_char *)(shared_qf_memory + CLIENT_REQUEST_FLAG_AREA) &=  ~0x40;
  }
+#endif
 
   if (not_use_sigio)
     io_select(fds_no_sigio);
 
-  /* this should be for per-second activities, it is actually at
-   * 180ms more or less */
-  partials++;
-  /* if you want a REAL second here use: config.freq*TIMER_DIVISOR */
-  if (partials == config.freq) {
-    partials = 0;
-#ifdef IPX
-  if (config.ipxsup)
-    pic_request (PIC_IPX);
-#endif
-    printer_tick((u_long) 0);
-    if (config.fastfloppy)
-      floppy_tick();
-  }
-
   /* Here we 'type in' prestrokes from commandline, as long as there are any
    * Were won't overkill dosemu, hence we type at a speed of 14cps
    */
   if (config.pre_stroke) {
-    static count=-1;
+    static int count=-1;
     if (--count < 0) {
       count = type_in_pre_strokes();
       if (count <0) count =7; /* with HZ=100 we have a stroke rate of 14cps */
     }
   }
 
+  /* this should be for per-second activities, it is actually at
+   * 200ms more or less (PARTIALS=5) */
+  if ((pic_sys_time-cnt200) >= (PIT_TICK_RATE/PARTIALS)) {
+    cnt200 = pic_sys_time;
+/*    g_printf("**** ALRM: %dms\n",(1000/PARTIALS)); */
+
+#ifdef IPX
+    if (config.ipxsup)
+      pic_request (PIC_IPX);
+#endif
+    printer_tick(0);
+    if (config.fastfloppy)
+      floppy_tick();
+  }
+
+/* We update the RTC from here if it has not been defined as a thread */
+
+  /* this is for EXACT per-second activities (can produce bursts) */
+  if ((pic_sys_time-cnt1000) >= PIT_TICK_RATE) {
+    cnt1000 += PIT_TICK_RATE;
+/*    g_printf("**** ALRM: 1sec\n"); */
+    rtc_update();
+  }
+
 }
 
 /* DANG_BEGIN_FUNCTION SIGNAL_save
@@ -431,6 +483,7 @@ void SIGALRM_call(void){
 inline void SIGNAL_save( void (*signal_call)() ) {
   signal_queue[SIGNAL_tail].signal_handler=signal_call;
   SIGNAL_tail = (SIGNAL_tail + 1) % MAX_SIG_QUEUE_SIZE;
+  signal_pending = 1;
   if (in_dpmi)
     dpmi_eflags |= VIP;
   REG(eflags) |= VIP;
@@ -464,25 +517,6 @@ sigio(int sig, struct sigcontext_struct context)
 }
 #endif
 
-#ifdef __NetBSD__
-#include <setjmp.h>
-
-extern sigjmp_buf handlerbuf;
-
-void
-sigio(int sig, int code, struct sigcontext *scp)
-{
-#ifdef DPMI
-  if (in_dpmi && !in_vm86)
-    dpmi_sigio(scp);
-#endif /* DPMI */
-  SIGNAL_save(SIGIO_call);
-    if (scp->sc_eflags & PSL_VM) {
-	vm86s.substr.regs.vmsc = *scp;
-	siglongjmp(handlerbuf, VM86_SIGNAL | 0x80000000);
-    }
-}
-#endif
 
 #ifdef __linux__
 void
@@ -492,24 +526,9 @@ sigalrm(int sig, struct sigcontext_struct context)
     dpmi_sigio(&context);
   SIGNAL_save(SIGALRM_call);
 }
+
 #endif
 
-#ifdef __NetBSD__
-void
-sigalrm(int sig, int code, struct sigcontext *scp)
-{
-#ifdef DPMI
-  if (in_dpmi && !in_vm86)
-    dpmi_sigio(scp);
-#endif /* DPMI */
-  h_printf("ding dong\n");
-  SIGNAL_save(SIGALRM_call);
-    if (scp->sc_eflags & PSL_VM) {
-	vm86s.substr.regs.vmsc = *scp;
-	siglongjmp(handlerbuf, VM86_SIGNAL | 0x80000000);
-    }
-}
-#endif
 
 void
 sigquit(int sig)
@@ -517,12 +536,12 @@ sigquit(int sig)
   in_vm86 = 0;
   in_sighandler = 1;
 
-  error("ERROR: sigquit called\n");
+  error("sigquit called\n");
   show_ints(0, 0x33);
   show_regs(__FILE__, __LINE__);
 
   ignore_segv++;
-  *(unsigned char *) 0x471 = 0x80;	/* ctrl-break flag */
+  *(unsigned char *)BIOS_KEYBOARD_FLAGS = 0x80;	/* ctrl-break flag */
   ignore_segv--;
 
   do_soft_int(0x1b);
