@@ -1,11 +1,11 @@
 /* 
- * (C) Copyright 1992, ..., 2000 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
 
 #define EXTERN 
-#define INIT(x...)		=  ## x
+#define INIT(x...)		=  x
 
 #include <termios.h>
 #include <sys/types.h>
