@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* 
  ******** OLD_DMA_CODE COPYRIGHT
  * dma.c--dma controller emulater--Joel N. Weber II
@@ -29,6 +35,8 @@
  * 17/3/97 - AM - Added in support for the handler to reject the completion.
  * 	    which allows the SB driver to work better, and better handling
  *	    when files "disappear" (EBADF) in some functions.
+ *         - AM - Added in the Auto-Init code.
+ * 4/7/97  - AM - Merging in Michael Karcher's code (and amending)
  * END_CHANGELOG
  */
 
@@ -45,26 +53,34 @@
 
 #include <unistd.h>
 
-/* AM - Inverted order of msb/lsb as CRISK found they were inverted. */
+/* #define EXCESSIVE_DEBUG 1 */
+
 typedef struct {
-  Bit8u msb;
   Bit8u lsb;
+  Bit8u msb;
+  Bit8u underflow;
 } msblsb_t;
 
 typedef union {
-  Bit8u data[2];
+  Bit8u data[3];
   msblsb_t bits;
   /* AM - Removed 'value' since this makes endian-ness assumptions */
 } multi_t;
 
 typedef struct {
+  int     mask;          /* Whether channel is masked */
   int     wfd;           /* Write part of pipe */
   int     rfd;           /* Read part of pipe */
   int     dack;          /* DACK */
   int     run;           /* Running */
   int     eop;           /* End Of Process */
   int     size;          /* Preferred Transfer Size */
-  int     (* handler)(int);  /* Handler Function */
+  int     (* handler)(int, Bit16u);  /* Handler Function */
+
+  multi_t address;       /* Backup address - for Auto-init modes */
+  multi_t length;        /* Backup length - for Auto-init modes */
+  int     set_size;      /* Backup transfer size - for Auto-init modes */
+
 } internal_t;
 
 typedef struct {
@@ -112,6 +128,7 @@ inline Bit16u get_value ( multi_t data );
 inline void add_value ( multi_t *data, Bit16u value );
 inline void sub_value ( multi_t *data, Bit16u value );
 inline void set_value ( multi_t *data, Bit16u value );
+inline int has_underflow ( multi_t data );
 
 inline int get_ch (int controller, int channel);
 inline int get_mask (int ch);
@@ -119,6 +136,8 @@ inline int get_mask (int ch);
 size_t dma_do_read(int controller, int channel, Bit32u target_addr);
 size_t dma_do_write(int controller, int channel, Bit32u target_addr);
 
+static inline void activate_channel (int controller, int channel);
+static inline void deactivate_channel (int controller, int channel);
 
 /* PUBLIC CODE */
 
@@ -147,10 +166,23 @@ void dma_assert_DREQ(int channel)
 
   controller = channel & 4;
   ch = channel & DMA_CH_SELECT;
-
   mask = 1 << (ch + 4);
+  
+#ifdef EXCESSIVE_DEBUG
+  h_printf ("DMA: ... mask = %d, status = %d\n", mask, 
+	    dma[controller].i[ch].mask);
+#endif /* EXCESSIVE_DEBUG */
+
+  /* 
+   * Must assert DREQ before activating the channel, otherwise it ignores the
+   * the request - AM
+   */
 
   dma[controller].status |= mask;
+
+  if(!dma[controller].i[ch].mask)
+    activate_channel (controller, ch);
+
 }
 
 void dma_drop_DACK(int channel)
@@ -179,6 +211,8 @@ void dma_drop_DREQ(int channel)
   controller = channel & 4;
   ch = channel & DMA_CH_SELECT;
 
+  deactivate_channel(controller, ch);
+
   mask = 1 << (ch +4);
   dma[controller].status &= ~ mask;
 }
@@ -268,7 +302,7 @@ int dma_test_eop(int channel)
 
 inline Bit16u get_value (multi_t data) 
 {
-	return ( ((Bit16u)data.bits.msb << 8) + data.bits.lsb);
+	return ( ((Bit16u)data.bits.msb << 8) | data.bits.lsb);
 }
 
 inline void add_value (multi_t *data, Bit16u value) 
@@ -279,7 +313,7 @@ inline void add_value (multi_t *data, Bit16u value)
 	h_printf ("DMA: MSB %u, LSB %u\n", data->bits.msb, data->bits.lsb);
 #endif /* EXCESSIVE_DEBUG */
 
-	tmp = ( ((Bit16u)data->bits.msb << 8) + data->bits.lsb);
+	tmp = ( ((Bit16u)data->bits.msb << 8) | data->bits.lsb);
 
 #ifdef EXCESSIVE_DEBUG
 	h_printf ("DMA: Adding %u to %u gives ", value, tmp);
@@ -293,6 +327,8 @@ inline void add_value (multi_t *data, Bit16u value)
 
 	data->bits.msb = (Bit8u) (tmp >> 8);
 	data->bits.lsb = (Bit8u) tmp & 0xFF;
+	/* ASSUMPTION: We added a positive number ... */
+	data->bits.underflow = 0;
 
 #ifdef EXCESSIVE_DEBUG
 	h_printf ("DMA: MSB %u, LSB %u\n", data->bits.msb, data->bits.lsb);
@@ -301,13 +337,13 @@ inline void add_value (multi_t *data, Bit16u value)
 
 inline void sub_value (multi_t *data, Bit16u value) 
 {
-	Bit16u tmp;
+	Bit32s tmp;
 
 #ifdef EXCESSIVE_DEBUG
 	h_printf ("DMA: MSB %u, LSB %u\n", data->bits.msb, data->bits.lsb);
 #endif /* EXCESSIVE_DEBUG */
 
-	tmp = ( ((Bit16u)data->bits.msb << 8) + data->bits.lsb);
+	tmp = (Bit32s) ( ((Bit16u)data->bits.msb << 8) | data->bits.lsb);
 
 #ifdef EXCESSIVE_DEBUG
 	h_printf ("DMA: Subtracting %u from %u gives ", value, tmp);
@@ -319,8 +355,15 @@ inline void sub_value (multi_t *data, Bit16u value)
 	h_printf ("%u\n", tmp);
 #endif /* EXCESSIVE_DEBUG */
 
-	data->bits.msb = (Bit8u) (tmp >> 8);
-	data->bits.lsb = (Bit8u) tmp & 0xFF;
+	if (tmp < 0) {
+	  data->bits.msb = 0;
+	  data->bits.lsb = 0;
+	  data->bits.underflow = 1;
+	} else {
+	  data->bits.msb = (Bit8u) (tmp >> 8);
+	  data->bits.lsb = (Bit8u) tmp & 0xFF;
+	  data->bits.underflow = 0;
+	}
 
 #ifdef EXCESSIVE_DEBUG
 	h_printf ("DMA: MSB %u, LSB %u\n", data->bits.msb, data->bits.lsb);
@@ -331,8 +374,56 @@ inline void set_value (multi_t *data, Bit16u value)
 {
 	data->bits.msb = (Bit8u) (value >> 8);
 	data->bits.lsb = (Bit8u) value & 0xFF;
+	data->bits.underflow = 0;
 }
 
+inline int has_underflow (multi_t data)
+{
+  return data.bits.underflow;
+}
+
+/*
+ * This ensures that the controller will be called to service this channel.
+ */
+static inline void activate_channel (int controller, int channel)
+{
+  int mask, ch;
+
+  ch = get_ch (controller, channel);
+
+  dma[controller].i[channel].mask = 0;
+  /* 
+   * We only _actually_ activate the channel if DREQ is set. It's irrelevant
+   * otherwise (from an idea by Michael Karcher) - AM
+   */
+     
+  if (dma_test_DREQ(ch) ) {
+    mask = get_mask (ch);
+
+    is_dma |= mask;
+#ifdef EXCESSIVE_DEBUG
+    h_printf ("DMA: Channel %d activated. DMA indicator is %d\n", ch, is_dma);
+#endif /* EXCESSIVE_DEBUG */
+  }
+}
+
+/* 
+ * This ensures that the controller is not called for this channel.
+ */
+static inline void deactivate_channel (int controller, int channel)
+{
+  int mask, ch;
+
+  dma[controller].i[channel].mask = 1;
+  ch = get_ch (controller, channel);
+  mask = get_mask (ch);
+
+  is_dma &= ~mask;  
+
+#ifdef EXCESSIVE_DEBUG
+  h_printf ("DMA: Channel %d deactivated. DMA indicator is %d\n", ch, is_dma);
+#endif /* EXCESSIVE_DEBUG */
+}
 
 inline Bit32u create_addr(Bit8u page, multi_t address, int dma_c)
 {
@@ -388,8 +479,37 @@ inline Bit8u dma_read_count (int dma_c, int channel)
   return r;
 }
 
+void dma_recover_values (int controller, int channel)
+{
+  extern dma_t dma[2];
 
-Bit8u dma_io_read(Bit32u port)
+  h_printf ("DMA: Recovering settings for Auto-Init transfer on %d, %d\n",
+	    controller, channel);
+  /* Recover the values from the private store */
+  dma[controller].address[channel].data[0] 
+    = dma[controller].i[channel].address.data[0];
+  dma[controller].address[channel].data[1]
+    = dma[controller].i[channel].address.data[1];
+  dma[controller].length[channel].data[0]
+    = dma[controller].i[channel].length.data[0];
+  dma[controller].length[channel].data[1]
+    = dma[controller].i[channel].length.data[1];
+  
+  dma[controller].i[channel].size = dma[controller].i[channel].set_size;
+}
+
+/* 
+ * Obselete? 
+ * Might make a useful public function if converted to just take a single
+ * parameter - AM 
+ */
+Bit16u length_transferred (int controller, int channel)
+{
+  return (get_value (dma[controller].i[channel].length)
+	  - get_value (dma[controller].length[channel]) );
+}
+
+Bit8u dma_io_read(ioport_t port)
 {
   Bit8u r;
 
@@ -467,6 +587,8 @@ Bit8u dma_io_read(Bit32u port)
 inline void dma_write_addr (int dma_c, int channel, Bit8u value)
 {
   dma[dma_c].address[channel].data[dma[dma_c].ff] = value;
+  dma[dma_c].i[channel].address.data[dma[dma_c].ff] = value; /* autoinit */
+
   h_printf ("DMA: Wrote 0x%x into Channel %d Address (Byte %d)\n", value,
 	    (dma_c *4) + channel, dma[dma_c].ff);
 
@@ -476,6 +598,8 @@ inline void dma_write_addr (int dma_c, int channel, Bit8u value)
 inline void dma_write_count (int dma_c, int channel, Bit8u value)
 {
   dma[dma_c].length[channel].data[dma[dma_c].ff] = value;
+  dma[dma_c].i[channel].length.data[dma[dma_c].ff] = value; /* autoinit */
+
   h_printf ("DMA: Wrote 0x%x into Channel %d Length (Byte %d)\n", value,
 	    (dma_c *4) + channel, dma[dma_c].ff);
 
@@ -487,15 +611,20 @@ inline void dma_write_mask (int dma_c, Bit8u value)
     if (value & DMA_SELECT_BIT)
     {
       dma[dma_c].current_channel = (int) value & 3;
+
+      deactivate_channel (dma_c, dma[dma_c].current_channel); /* - Karcher */
+
       h_printf ("DMA: Channel %u selected.\n", 
 			(dma_c * 4) + dma[dma_c].current_channel);
+      /* Clear the COMPLETE flag for this channel */
+      dma[dma_c].status &= ~ (1 << (value & 3));
     }
     else
     {
       h_printf ("DMA: Channel %u deselected.\n", 
-			 (dma_c * 4) + dma[dma_c].current_channel);
+			 (dma_c * 4) + (value&3));
 
-      is_dma |= (1 << dma[dma_c].current_channel);
+      activate_channel (dma_c, dma[dma_c].current_channel);
 
       dma[dma_c].current_channel = -1;
     }
@@ -504,7 +633,7 @@ inline void dma_write_mask (int dma_c, Bit8u value)
 
 
 
-void dma_io_write(Bit32u port, Bit8u value)
+void dma_io_write(ioport_t port, Bit8u value)
 {
   Bit8u ch;
 
@@ -619,12 +748,14 @@ void dma_io_write(Bit32u port, Bit8u value)
   case DMA1_CLEAR_FF_REG:
     h_printf ("DMA: Clearing DMA1 Output FF\n");
                 /* Kernel implies this, then ignores it */
-    dma[DMA1].ff = value & 1;
+                /* dma[DMA1].ff = value & 1; */
+    dma[DMA1].ff = 0;
     break;
   case DMA2_CLEAR_FF_REG:
     h_printf ("DMA: Clearing DMA2 Output FF\n");
                 /* Kernel implies this, then ignores it */
-    dma[DMA2].ff = value & 1;
+                /* dma[DMA2].ff = value & 1; */
+    dma[DMA2].ff = 0;
     break;
 
   default:
@@ -634,8 +765,8 @@ void dma_io_write(Bit32u port, Bit8u value)
 
 
 
-void dma_install_handler (int ch, int wfd, int rfd, int (* handler) (int), 
-			  int size)
+void dma_install_handler (int ch, int wfd, int rfd, 
+			  int (* handler) (int, Bit16u), int size)
 {
   int channel, dma_c;
 
@@ -645,32 +776,34 @@ void dma_install_handler (int ch, int wfd, int rfd, int (* handler) (int),
   dma_c = ch & 4;
   channel = ch & DMA_CH_SELECT;
 
-  if (dma[dma_c].i[channel].wfd != -1)
+  if (dma[dma_c].i[channel].wfd != -1 && dma[dma_c].i[channel].wfd != wfd)
     close (dma[dma_c].i[channel].wfd);
   dma[dma_c].i[channel].wfd = wfd;
 
-  if (dma[dma_c].i[channel].rfd != -1)
+  if (dma[dma_c].i[channel].rfd != -1 && dma[dma_c].i[channel].rfd != rfd)
     close (dma[dma_c].i[channel].rfd);
   dma[dma_c].i[channel].rfd = rfd;
 
   dma[dma_c].i[channel].handler = handler;
 
   dma[dma_c].i[channel].size = size;
+  dma[dma_c].i[channel].set_size = size;
 }
 
 
 /*
- * This sets up a DMA channel for the first time.
+ * This sets up a DMA channel when it used for the first time after 
+ * programming ie if there are no files alreadu associated with it.
  */
 
 int dma_initialise_channel (int controller, int channel) 
 {
 	extern dma_t dma[2];
 	int tmp_pipe[2];
-	int ch, mask;
+	int ch/*, mask*/;
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch); */
 
 	h_printf ("DMA: Initialising transfer for channel %d, controller %d\n",
 		  ch, controller +1);
@@ -684,7 +817,9 @@ int dma_initialise_channel (int controller, int channel)
 		 * DANG_FIXTHIS: Pipe Creation failed. Lets hope that it times out ! 
 		 */
 
-		is_dma &= ~mask;
+	  /*is_dma &= ~mask; */
+	        deactivate_channel (controller, channel);
+
 		return -1;
 	}
 
@@ -700,13 +835,13 @@ int dma_initialise_channel (int controller, int channel)
 	 * DANG_END_REMARK
 	 */
 
-/*      Why Comment this out ? - AM
-	if (dma[controller].ch_mode[channel] & (DMA_AUTO_INIT | DMA_READ))
+	/*      Why Comment this out ? - AM */
+	if (dma[controller].ch_mode[channel] & (DMA_AUTO_INIT | DMA_READ)) {
 	  dma_assert_DREQ (ch);
-	else
- */                       
-	h_printf("DMA: [crisk] dropping DREQ\n");
-	dma_drop_DREQ(ch);
+	} else {
+	  h_printf("DMA: [crisk] dropping DREQ\n");
+	  dma_drop_DREQ(ch);
+	}
 
 	return 0;
 }
@@ -749,38 +884,55 @@ void dma_process_demand_mode_write (int controller, int channel)
 	Bit32u target_addr;
 	int amount_done = 0;
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
       target_addr = create_addr (dma[controller].page[channel],
 				  dma[controller].address[channel],
 				 controller);
+#ifdef EXCESSIVE_DEBUG
+      h_printf ("DMA: Demand Mode Write run (%d) target: %u\n", ch,
+		target_addr);
+#endif /* EXCESSIVE_DEBUG */
 
 	if (dma_test_DACK(ch)) {
 	    dma_drop_DACK(ch);
 	} else {
+#ifdef EXCESSIVE_DEBUG
+	    h_printf ("DMA: Demand Mode Write run aborted. DACK not set\n");
+#endif /* EXCESSIVE_DEBUG */
 	    return;
 	}
 
-	if (!get_value (dma[controller].length[channel])) {
-	    /* Transfer is complete */ 
+	if (!get_value (dma[controller].length[channel])
+	    && has_underflow(dma[controller].length[channel])) {
+	  /* Transfer is complete */ 
+	  if ( dma[controller].ch_mode[channel] & DMA_AUTO_INIT ) {
+	    /* Auto-Init means re-start, so we re-start not stop */
+	    if (dma[controller].i[channel].handler != NULL) {
+	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE, 0);
+	    }
+
+	    dma_recover_values (controller, channel);
+	  } else {
 	    dma[controller].status |= (1 << channel);
 
-		is_dma &= ~ (1 << (ch -1));
-				
-		if (dma[controller].i[channel].handler != NULL) {
-			/* The handler is expected to close any descriptors */
-	      dma[controller].i[channel].handler (DMA_HANDLER_DONE);
-		} else {
-			close (dma[controller].i[channel].wfd);
-		}
-		
-		dma[controller].i[channel].wfd = -1; 
+	    /*is_dma &= ~ (1 << (ch -1));*/
+	    deactivate_channel (controller, channel);
 
-	    return; 
+	    if (dma[controller].i[channel].handler != NULL) {
+	      /* The handler is expected to close any descriptors */
+	      dma[controller].i[channel].handler (DMA_HANDLER_DONE, 0);
+	    } else {
+	      close (dma[controller].i[channel].wfd);
+	    }
+		
+	    dma[controller].i[channel].wfd = -1; 
 	  }
+	  return; 
+	}
 
 	if (dma_test_eop(ch)) {
 	    /* Stop the Transfer */
@@ -790,17 +942,18 @@ void dma_process_demand_mode_write (int controller, int channel)
 	    dma[controller].i[channel].wfd = -1;
 	    */ 
 		
-		is_dma &= ~mask;
+	    deactivate_channel (controller, channel);
+		/*		is_dma &= ~mask;*/
 	    return;
 	  }
 
 	if (dma_test_DREQ(ch)) {
 	    dma[controller].i[channel].run = 1; /* TRIVIAL */
 
-		if (get_value (dma[controller].length[channel])
+		if (get_value (dma[controller].length[channel]) +1
 		    < dma[controller].i[channel].size) {
 	      dma[controller].i[channel].size 
-				= get_value (dma[controller].length[channel]);
+			      = get_value (dma[controller].length[channel]) +1;
 		}
  	    amount_done = dma_do_read(controller, channel, target_addr);
 
@@ -815,7 +968,8 @@ void dma_process_demand_mode_write (int controller, int channel)
 		sub_value (&dma[controller].length[channel], amount_done);
 
 		if (dma[controller].i[channel].handler != NULL) {
-	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE);
+	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE,
+						  amount_done);
 	  }
 	}
 }
@@ -825,24 +979,31 @@ void dma_process_single_mode_write (int controller, int channel)
 	Bit32u target_addr;
 	int amount_done = 0;
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	target_addr = create_addr(dma[controller].page[channel],
 				  dma[controller].address[channel],
 				  controller);
+
+#ifdef EXCESSIVE_DEBUG
+      h_printf ("DMA: single Mode Write run (%d) target: %u\n", ch, 
+		target_addr);
+#endif /* EXCESSIVE_DEBUG */
 	
-	if (! get_value (dma[controller].length[channel]) ) {
+	if (! get_value (dma[controller].length[channel]) 
+	    && has_underflow(dma[controller].length[channel])) {
 	    /* Transfer is complete */
 	    dma[controller].status |= (1 << channel);
 
-		is_dma &= ~mask;
+	        deactivate_channel (controller, channel);
+		/*		is_dma &= ~mask;*/
 				
 		if (dma[controller].i[channel].handler != NULL) {
 			/* The handler is expected to close any descriptors */
-	      dma[controller].i[channel].handler (DMA_HANDLER_DONE);
+	      dma[controller].i[channel].handler (DMA_HANDLER_DONE, 0);
 		} else {			
 			close (dma[controller].i[channel].wfd);
 		}
@@ -860,10 +1021,10 @@ void dma_process_single_mode_write (int controller, int channel)
 	      dma_drop_DACK(ch);
 		}
 
-		if (get_value (dma[controller].length[channel])
+		if (get_value (dma[controller].length[channel]) +1
 		    < dma[controller].i[channel].size) {
 	      dma[controller].i[channel].size 
-				= get_value (dma[controller].length[channel]);
+			      = get_value (dma[controller].length[channel]) +1;
 		}
 
  	    amount_done = dma_do_read(controller, channel, target_addr);
@@ -879,7 +1040,8 @@ void dma_process_single_mode_write (int controller, int channel)
 		sub_value (&dma[controller].length[channel], amount_done);
 
 		if (dma[controller].i[channel].handler != NULL) {
-	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE);
+	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE,
+						  amount_done);
 	  }
 	}
 }	
@@ -890,39 +1052,55 @@ void dma_process_block_mode_write (int controller, int channel)
 	Bit32u target_addr;
 	int amount_done = 0;
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	target_addr = create_addr(dma[controller].page[channel],
 				  dma[controller].address[channel],
 				  controller);
 	
+#ifdef EXCESSIVE_DEBUG
+      h_printf ("DMA: block Mode Write run (%d) target: %u\n", ch, 
+		target_addr);
+#endif /* EXCESSIVE_DEBUG */
+
 	if (dma_test_DACK(ch)) {
 	    dma_drop_DACK(ch);
 	} else {
 	    return;
 	}
 
-	if (! get_value (dma[controller].length[channel]) ) {
-	    /* Transfer is complete */
+	if (! get_value (dma[controller].length[channel]) 
+	    && has_underflow(dma[controller].length[channel])) {
+	  /* Transfer is complete */
+	  if ( dma[controller].ch_mode[channel] & DMA_AUTO_INIT ) {
+	    /* Auto-Init means re-start, so we re-start not stop */
+	    if (dma[controller].i[channel].handler != NULL) {
+	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE, 0);
+	    }
+
+	    dma_recover_values (controller, channel);
+	  } else {
 	    dma[controller].status |= (1 << channel);
 
-		is_dma &= ~mask;
+	    deactivate_channel (controller, channel);
+		/*	    is_dma &= ~mask;*/
 		
-		if (dma[controller].i[channel].handler != NULL) {
-			/* The handler is expected to close any descriptors */
-	      dma[controller].i[channel].handler (DMA_HANDLER_DONE);
-		} else {
-			close (dma[controller].i[channel].wfd);
-		}
+	    if (dma[controller].i[channel].handler != NULL) {
+	      /* The handler is expected to close any descriptors */
+	      dma[controller].i[channel].handler (DMA_HANDLER_DONE,0);
+	    } else {
+	      close (dma[controller].i[channel].wfd);
+	    }
 
-		dma[controller].i[channel].wfd = -1;
-
-	    return;
+	    dma[controller].i[channel].wfd = -1;
 	  }
 
+	  return;
+	}
+
 	if (dma_test_eop(ch)) {
 	    /* Stop the Transfer */
 	    dma[controller].status &= (1 << channel);
@@ -930,7 +1108,8 @@ void dma_process_block_mode_write (int controller, int channel)
 /*	    close (dma[controller].i[channel].wfd);
 	    dma[controller].i[channel].wfd = -1;
 	    */ 
-		is_dma &= ~mask;
+	        deactivate_channel (controller, channel);
+		/*		is_dma &= ~mask;*/
 	    return;
 	  }
 
@@ -938,10 +1117,10 @@ void dma_process_block_mode_write (int controller, int channel)
 	    dma[controller].i[channel].run = 1; /* Important ! */
 
 	if (dma[controller].i[channel].run) {
-		if ( get_value (dma[controller].length[channel])
+		if ( get_value (dma[controller].length[channel]) +1
 		     < dma[controller].i[channel].size) {
 	      dma[controller].i[channel].size 
-				= get_value (dma[controller].length[channel]);
+			      = get_value (dma[controller].length[channel]) +1;
 		}
 
  	    amount_done = dma_do_read(controller, channel, target_addr);
@@ -957,22 +1136,23 @@ void dma_process_block_mode_write (int controller, int channel)
 		sub_value (&dma[controller].length[channel], amount_done);
 	    
 		if (dma[controller].i[channel].handler != NULL) {
-	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE);
+	      dma[controller].i[channel].handler (DMA_HANDLER_WRITE,
+						  amount_done);
 	  }
 	}
 }
 
 
 /* 
- * DANG_FIXTHIS: Cascade mode Writes are not supported 
+ * Cascade mode Writes are not supported - Unimportant
  */
 void dma_process_cascade_mode_write (int controller, int channel)
 {
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	  h_printf ("DMA: Attempt to use unsupported CASCADE mode\n");
 
@@ -980,10 +1160,11 @@ void dma_process_cascade_mode_write (int controller, int channel)
 	  dma[controller].i[channel].wfd = -1;
 	  */ 
 
-	is_dma &= ~mask;
+	  deactivate_channel (controller, channel);
+	  /*	is_dma &= ~mask;*/
 
 	if (dma[controller].i[channel].handler != NULL) {
-		dma[controller].i[channel].handler(DMA_HANDLER_DONE);
+		dma[controller].i[channel].handler(DMA_HANDLER_DONE, 0);
 	}
 }
 
@@ -993,14 +1174,19 @@ void dma_process_demand_mode_read (int controller, int channel)
 	Bit32u target_addr;
 	int amount_done = 0;
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	target_addr = create_addr(dma[controller].page[channel],
 				  dma[controller].address[channel],
 				  controller);
+
+#ifdef EXCESSIVE_DEBUG
+      h_printf ("DMA: demand Mode read run (%d) target: %u\n", ch, 
+		target_addr);
+#endif /* EXCESSIVE_DEBUG */
 	
 	if (dma_test_DACK(ch)) {
 	    dma_drop_DACK(ch);
@@ -1008,23 +1194,33 @@ void dma_process_demand_mode_read (int controller, int channel)
 	    return;
 	}
 
-	if (! get_value(dma[controller].length[channel]) ) {
-	    /* Transfer is complete */
+	if (! get_value(dma[controller].length[channel]) 
+	    && has_underflow(dma[controller].length[channel])) {
+	  /* Transfer is complete */
+	  if ( dma[controller].ch_mode[channel] & DMA_AUTO_INIT ) {
+	    /* Auto-Init means re-start, so we re-start not stop */
+	    if (dma[controller].i[channel].handler != NULL) {
+	      dma[controller].i[channel].handler (DMA_HANDLER_READ, 0);
+	    }
+
+	    dma_recover_values (controller, channel);
+	  } else {
 	    dma[controller].status |= (1 << channel);
 
-		is_dma &= ~mask;
-				
-		if (dma[controller].i[channel].handler != NULL) {
-			/* The handler is expected to close any descriptors */
-	      dma[controller].i[channel].handler (DMA_HANDLER_DONE);
-		} else {
-			close (dma[controller].i[channel].rfd);
-		}
-
-		dma[controller].i[channel].rfd = -1;
-
-	    return;
+	    deactivate_channel (controller, channel);
+	    /*is_dma &= ~mask;*/
+	    
+	    if (dma[controller].i[channel].handler != NULL) {
+	      /* The handler is expected to close any descriptors */
+	      dma[controller].i[channel].handler (DMA_HANDLER_DONE, 0);
+	    } else {
+	      close (dma[controller].i[channel].rfd);
+	    }
+	    
+	    dma[controller].i[channel].rfd = -1;
 	  }
+	  return;
+	}
 
 	if (dma_test_eop(ch)) {
 	    /* Stop the Transfer */
@@ -1033,17 +1229,18 @@ void dma_process_demand_mode_read (int controller, int channel)
 /*	    close (dma[controller].i[channel].rfd);
 	    dma[controller].i[channel].rfd = -1;
 	    */ 
-		is_dma &= ~mask;
+	    deactivate_channel (controller, channel);
+	    /*is_dma &= ~mask;*/
 	    return;
 	  }
 
 	if (dma_test_DREQ(ch)) {
 	    dma[controller].i[channel].run = 1; /* TRIVIAL */
 
-		if (get_value (dma[controller].length[channel])
+		if (get_value (dma[controller].length[channel]) +1
 		    < dma[controller].i[channel].size) {
 	      dma[controller].i[channel].size 
-				= get_value (dma[controller].length[channel]);
+			      = get_value (dma[controller].length[channel]) +1;
 		}
 
  	    amount_done = dma_do_write(controller, channel, target_addr);
@@ -1060,7 +1257,8 @@ void dma_process_demand_mode_read (int controller, int channel)
 			   amount_done);
 
 		if (dma[controller].i[channel].handler != NULL) {
-	      dma[controller].i[channel].handler (DMA_HANDLER_READ);
+	      dma[controller].i[channel].handler (DMA_HANDLER_READ,
+						  amount_done);
 		}
 	}
 }
@@ -1071,42 +1269,52 @@ void dma_process_single_mode_read (int controller, int channel)
 	Bit32u target_addr;
 	size_t amount_done = 0;
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	target_addr = create_addr(dma[controller].page[channel],
 				  dma[controller].address[channel],
 				  controller);
 	
 	h_printf("DMA: Single Mode Read - length %d (%d)\n", 
-		 get_value (dma[controller].length[channel]),
+		 get_value (dma[controller].length[channel]) +1,
 		 dma[controller].i[channel].size);
 
-	if (!get_value (dma[controller].length[channel])) {
-	    /* Transfer is complete */
-		h_printf("DMA: [crisk] Transfer is complete\n");    
-		if (dma[controller].i[channel].handler != NULL) {
-			/* The handler is expected to close any descriptors */
-			if ((*dma[controller].i[channel].handler)(DMA_HANDLER_DONE) 
-			    != DMA_HANDLER_OK) {
-			  h_printf ("DMA: Handler indicates incomplete.\n");
-			  return;
-			} else {
-			  h_printf ("DMA: Handler indicates complete.\n");
-			}
-		} else {
-			close (dma[controller].i[channel].rfd);
-	  }
+	if (!get_value (dma[controller].length[channel])
+	    && has_underflow(dma[controller].length[channel])) {
+	  /* Transfer is complete */
+	  h_printf("DMA: [crisk] Transfer is complete\n");    
+	  if ( dma[controller].ch_mode[channel] & DMA_AUTO_INIT ) {
+	    /* Auto-Init means re-start, so we re-start not stop */
+	    if (dma[controller].i[channel].handler != NULL) {
+	      dma[controller].i[channel].handler (DMA_HANDLER_READ, 0);
+	    }
+	    
+	    dma_recover_values (controller, channel);
+	  } else {
+	    if (dma[controller].i[channel].handler != NULL) {
+	      /* The handler is expected to close any descriptors */
+	      if ((*dma[controller].i[channel].handler)(DMA_HANDLER_DONE, 0) 
+		  != DMA_HANDLER_OK) {
+		h_printf ("DMA: Handler indicates incomplete.\n");
+		return;
+	      } else {
+		h_printf ("DMA: Handler indicates complete.\n");
+	      }
+	    } else {
+	      close (dma[controller].i[channel].rfd);
+	    }
 
 	    dma[controller].status |= (1 << channel);
 
-		is_dma &= ~mask;
+	    deactivate_channel (controller, channel);
+	    /*is_dma &= ~mask;*/
 				
-		dma[controller].i[channel].rfd = -1;
-				
-		return;
+	    dma[controller].i[channel].rfd = -1;
+	  }		
+	  return;
 	}
 			
 	h_printf("DMA: [crisk] SM read (%d && %d) || (%d && %d)\n",
@@ -1124,10 +1332,10 @@ void dma_process_single_mode_read (int controller, int channel)
 	      dma_drop_DACK(ch);
 		}
 
-		if (get_value (dma[controller].length[channel])
+		if (get_value (dma[controller].length[channel]) +1
 		    < dma[controller].i[channel].size) {
 	      dma[controller].i[channel].size 
-				= get_value (dma[controller].length[channel]);
+			      = get_value (dma[controller].length[channel]) +1;
 		}
 
 		h_printf("DMA: [crisk] calling dma_do_write()\n");
@@ -1145,7 +1353,8 @@ void dma_process_single_mode_read (int controller, int channel)
 		sub_value (&dma[controller].length[channel], amount_done);
 
 		if (dma[controller].i[channel].handler != NULL) {
-	      dma[controller].i[channel].handler (DMA_HANDLER_READ);
+	      dma[controller].i[channel].handler (DMA_HANDLER_READ,
+						  amount_done);
 	  }
 	}
 	h_printf("DMA: [crisk] DMA single read end trace\n");
@@ -1158,14 +1367,19 @@ void dma_process_block_mode_read (int controller, int channel)
 	Bit32u target_addr;
 	int amount_done = 0;
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	target_addr = create_addr(dma[controller].page[channel],
 				  dma[controller].address[channel],
 				  controller);
+
+#ifdef EXCESSIVE_DEBUG
+      h_printf ("DMA: block Mode read run (%d) target: %u\n", ch, 
+		target_addr);
+#endif /* EXCESSIVE_DEBUG */
 	
 	if (dma_test_DACK(ch)) {
 	    dma_drop_DACK(ch);
@@ -1173,23 +1387,33 @@ void dma_process_block_mode_read (int controller, int channel)
 	    return;
 	}
 
-	if (! get_value (dma[controller].length[channel]) ) {
-	    /* Transfer is complete */
+	if (! get_value (dma[controller].length[channel]) 
+	    && has_underflow(dma[controller].length[channel])) {
+	  /* Transfer is complete */
+	  if ( dma[controller].ch_mode[channel] & DMA_AUTO_INIT ) {
+	    /* Auto-Init means re-start, so we re-start not stop */
+	    if (dma[controller].i[channel].handler != NULL) {
+	      dma[controller].i[channel].handler (DMA_HANDLER_READ, 0);
+	    }
+
+	    dma_recover_values (controller, channel);
+	  } else {
 	    dma[controller].status |= (1 << channel);
 
-		is_dma &= ~mask;
+	    deactivate_channel (controller, channel);
+	    /*is_dma &= ~mask;*/
 
-		if (dma[controller].i[channel].handler != NULL) {
-			/* The handler is expected to close any descriptors */
-	      dma[controller].i[channel].handler (DMA_HANDLER_DONE);
-		} else {
-			close (dma[controller].i[channel].rfd);
-		}
-
-		dma[controller].i[channel].rfd = -1;
-
-	    return;
+	    if (dma[controller].i[channel].handler != NULL) {
+	      /* The handler is expected to close any descriptors */
+	      dma[controller].i[channel].handler (DMA_HANDLER_DONE, 0);
+	    } else {
+	      close (dma[controller].i[channel].rfd);
+	    }
+	    
+	    dma[controller].i[channel].rfd = -1;
 	  }
+	  return;
+	}
 
 	if (dma_test_eop(ch)) {
 	    /* Stop the Transfer */
@@ -1199,7 +1423,8 @@ void dma_process_block_mode_read (int controller, int channel)
 	    dma[controller].i[channel].rfd = -1;
 	    */ 
 
-		is_dma &= ~mask;
+	    deactivate_channel (controller, channel);
+	    /*is_dma &= ~mask;*/
 	    return;
 	  }
 
@@ -1207,11 +1432,11 @@ void dma_process_block_mode_read (int controller, int channel)
 	    dma[controller].i[channel].run = 1; /* Essential ! */
 
 	if (dma[controller].i[channel].run) {
-		if ( get_value (dma[controller].length[channel] )
+		if ( get_value (dma[controller].length[channel]) +1
 		     < dma[controller].i[channel].size) {
 
 	      dma[controller].i[channel].size 
-				= get_value (dma[controller].length[channel]);
+			      = get_value (dma[controller].length[channel]) +1;
 		}
 
  	    amount_done = dma_do_write(controller, channel, target_addr);
@@ -1227,7 +1452,8 @@ void dma_process_block_mode_read (int controller, int channel)
 		sub_value (&dma[controller].length[channel], amount_done);
 	    
 		if (dma[controller].i[channel].handler != NULL) {
-	      dma[controller].i[channel].handler (DMA_HANDLER_READ);
+	      dma[controller].i[channel].handler (DMA_HANDLER_READ,
+						  amount_done);
 	  }
 	}
 }
@@ -1239,17 +1465,19 @@ void dma_process_block_mode_read (int controller, int channel)
 void dma_process_cascade_mode_read (int controller, int channel)
 {
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
 	  h_printf ("DMA: Attempt to use unsupported CASCADE mode\n");
 
-	is_dma &= ~mask;
+	  /*is_dma &= ~mask;*/
+	deactivate_channel (controller, channel);
+
 
 	if (dma[controller].i[channel].handler != NULL) {
-	    dma[controller].i[channel].handler (DMA_HANDLER_DONE);
+	    dma[controller].i[channel].handler (DMA_HANDLER_DONE, 0);
 	} else {
 		close (dma[controller].i[channel].rfd);
 	}
@@ -1263,16 +1491,16 @@ void dma_process_cascade_mode_read (int controller, int channel)
 void dma_process_verify_mode  (int controller, int channel)
 {
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
  	h_printf ("DMA: Attempt to use unsupported VERIFY direction by controller %d, channel %d\n",
 		  controller +1, channel);
 	
 	if (dma[controller].i[channel].handler != NULL) {
-		dma[controller].i[channel].handler(DMA_HANDLER_DONE);
+		dma[controller].i[channel].handler(DMA_HANDLER_DONE,0);
 	} else {
 	close (dma[controller].i[channel].rfd);
 		close(dma[controller].i[channel].wfd);
@@ -1281,7 +1509,8 @@ void dma_process_verify_mode  (int controller, int channel)
 	dma[controller].i[channel].rfd = -1;
 	dma[controller].i[channel].wfd = -1;
 
-	is_dma &= ~mask;
+	deactivate_channel (controller, channel);
+	/*is_dma &= ~mask;*/
 }
 
 
@@ -1291,16 +1520,16 @@ void dma_process_verify_mode  (int controller, int channel)
 void dma_process_invalid_mode  (int controller, int channel)
 {
 	int ch;
-	int mask;
+	/*	int mask;*/
 
 	ch = get_ch (controller, channel);
-	mask = get_mask (ch);
+	/*	mask = get_mask (ch);*/
 
  	h_printf ("DMA: Attempt to use INVALID direction by controller %d, channel %d\n",
 		  controller +1, channel);
 
 	if (dma[controller].i[channel].handler != NULL) {
-		dma[controller].i[channel].handler(DMA_HANDLER_DONE);
+		dma[controller].i[channel].handler(DMA_HANDLER_DONE, 0);
 	} else {
 	close (dma[controller].i[channel].rfd);
 		close(dma[controller].i[channel].wfd);
@@ -1309,7 +1538,8 @@ void dma_process_invalid_mode  (int controller, int channel)
 	dma[controller].i[channel].rfd = -1;
 	dma[controller].i[channel].wfd = -1;
 
-	is_dma &= ~mask;
+	deactivate_channel (controller, channel);
+	/*is_dma &= ~mask;*/
 }
 
 
@@ -1389,6 +1619,21 @@ int dma_process_channel (int controller, int channel) {
     return 0;
 }
 
+/* Karcher */
+void dma_transfer_size (int ch, int size)
+{
+  int channel, dma_c;
+
+  h_printf ("DMA: Changing DMA block-size on channel %d [%d bytes/call]\n", 
+	    ch, size);
+
+  dma_c = ch & 4;
+  channel = ch & DMA_CH_SELECT;
+
+  dma[dma_c].i[channel].size = size;
+  dma[dma_c].i[channel].set_size = size; /* For auto-init - AM*/
+}
+
 
 /*
  * This is surrounds the brains of the operation ....
@@ -1399,12 +1644,19 @@ void dma_controller(void)
   Bit8u test;
   int controller, channel;
 
-  for (test = 1, controller = DMA1, channel = 0; test != 0; test = test << 1, channel++) {
+  for (test = 1, controller = DMA1, channel = 0; test > 0 && test <= 128; 
+       test = test << 1, channel++) {
     if (channel == 4) {
 	controller = DMA2;
 	channel = 0;
     }
 
+#ifdef EXCESSIVE_DEBUG
+    h_printf ("DMA: controller: (%d) run - %s, current - %s\n", 
+	      (controller * 4) + channel, (is_dma & test) ? "true": "false",
+	      (dma[controller].current_channel == channel) ? "true": "false");
+#endif /* EXCESSIVE_DEBUG */
+
 		/* Process the channel only if it has been deselected */
     if ((is_dma & test) && (dma[controller].current_channel != channel)) {
 	/* Time to process this DMA channel */
@@ -1432,7 +1684,7 @@ size_t dma_do_write(int controller, int channel, Bit32u target_addr)
     if (errno == EBADF) {
       h_printf("DMA: BAD File on Channel %d of controller %d (remaining length: %d)\n",
 				channel, controller, 
-				get_value(dma[controller].length[channel]));
+				get_value(dma[controller].length[channel])+1);
     set_value (&dma[controller].length[channel], 0);
   } else if (errno != EINTR) {
     h_printf ("DMA: Error in READ on Channel %d of controller %d (%s)\n", 
@@ -1470,21 +1722,28 @@ size_t dma_do_read(int controller, int channel, Bit32u target_addr)
     return 0;
   }
   else {
+#ifdef EXCESSIVE_DEBUG
+    h_printf ("DMA: Write %u bytes\n", ret_val);
+#endif /* EXCESSIVE_DEBUG */
+
     return ret_val;
   }
 }
 
 void dma_init(void)
 {
-  emu_iodev_t  io_device;
   int i, j;
+  emu_iodev_t  io_device;
   
   /* 8237 DMA controller */
   io_device.read_portb   = dma_io_read;
   io_device.write_portb  = dma_io_write;
   io_device.read_portw   = NULL;
   io_device.write_portw  = NULL;
+  io_device.read_portd   = NULL;
+  io_device.write_portd  = NULL;
   io_device.irq          = EMU_NO_IRQ;
+  io_device.fd           = -1;
 
   /*
    * XT Controller 
@@ -1492,7 +1751,7 @@ void dma_init(void)
   io_device.start_addr   = 0x0000;
   io_device.end_addr     = 0x000F;
   io_device.handler_name = "DMA - XT Controller";
-  port_register_handler(io_device);
+  port_register_handler(io_device, 0);
 
   /*
    * Page Registers (XT)
@@ -1500,7 +1759,7 @@ void dma_init(void)
   io_device.start_addr   = 0x0081;
   io_device.end_addr     = 0x0087;
   io_device.handler_name = "DMA - XT Pages";
-  port_register_handler(io_device);
+  port_register_handler(io_device, 0);
 
   /*
    * AT Controller
@@ -1508,7 +1767,7 @@ void dma_init(void)
   io_device.start_addr   = 0x00C0;
   io_device.end_addr     = 0x00DE;
   io_device.handler_name = "DMA - AT Controller";
-  port_register_handler(io_device);
+  port_register_handler(io_device, 0);
 
   /*
    * Page Registers (AT)
@@ -1516,10 +1775,10 @@ void dma_init(void)
   io_device.start_addr   = 0x0089;
   io_device.end_addr     = 0x008A;
   io_device.handler_name = "DMA - AT Pages";
-  port_register_handler(io_device);
+  port_register_handler(io_device, 0);
 
   for (i = 0; i < 2; i++) {
-    for (j = 0; j < 5; j++) {
+    for (j = 0; j < 4; j++) {
       set_value (&dma[i].address[j], 0);
       set_value (&dma[i].length[j], 0);
       dma[i].ch_config[j] = 0;
@@ -1541,7 +1800,7 @@ void dma_init(void)
 
   is_dma = 0;
 
-  h_printf ("DMA: DMA Controller initialiased - 8 & 16 bit modes\n");
+  h_printf ("DMA: DMA Controller initialized - 8 & 16 bit modes\n");
 }
   
 
