@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -142,9 +142,6 @@
 #include "pic.h"
 #include "iodev.h"
 #include "memory.h"
-#ifdef __linux__
-#include <linux/linkage.h>
-#endif
 /* #include <sys/vm86.h> */
 #include <sys/time.h>
 #include "cpu.h"
@@ -183,7 +180,7 @@ static unsigned long pic1_mask = 0x07f8; /* bits set for pic1 levels */
 even asgcc -Wall says;
 pic.c: At top level:
 pic.c:81: warning: `pic1_mask' defined but not used
-it it used in the assembly part
+though it _is_ used in the assembly part
 */
 static unsigned long   pic_smm = 0;      /* 32=>special mask mode, 0 otherwise */
 
@@ -200,15 +197,15 @@ unsigned long   pic_stack[32];      /* stack of interrupt levels */
 unsigned long   pic_sp = 0;         /* pointer to stack */
 unsigned long   pic_vm86_count=0;   /* counter for trips around vm86 loop */
 unsigned long   pic_dpmi_count=0;   /* counter for trips around dpmi loop */
-         long   pic_sys_time=NEVER; /* system time for scheduling interrupts */
+         hitimer_t pic_sys_time=NEVER; /* system time for scheduling interrupts */
 #endif
-         long   pic_ltime[33] =     /* timeof last pic request honored */
+static   hitimer_t pic_ltime[33] =     /* timeof last pic request honored */
                 {NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER};
-         long   pic_itime[33] =     /* time to trigger next interrupt */
+         hitimer_t pic_itime[33] =     /* time to trigger next interrupt */
                 {NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
@@ -343,7 +340,7 @@ char ci,cc;
 
  * DANG_END_REMARK pic_print
  */
-void inline pic_push(int val)
+inline void pic_push(int val)
 {
     if(pic_sp<32){
        pic_stack[pic_sp++]=val;
@@ -700,6 +697,17 @@ if ((!(REG(eflags) & VIF)) && (!in_dpmi) ) {
 #else
 if (!(REG(eflags) & VIF)) return;
 #endif
+
+/* Without this one dope crashes (DPMI_rm_procedure_* running increases 
+ * up to a stack fault)
+ *
+ * This is a gross disgusting cludge, that I am going to make even grosser
+ * and more disgusting in an attempt to get it to work properly.  Someday
+ * I am going to find and fix the root problems that are making this 
+ * necessary.  -KenC
+ */
+if (in_dpmi && in_dpmi_timer_int) return;
+
 #warning using assembly run_irqs,expect pic1_mask not used
   __asm__ __volatile__
   ("movl "CISH_INLINE(pic_ilevel)",%%ecx\n\t" /* get old ilevel              */
@@ -780,9 +788,9 @@ int do_irq()
     if(IS_REDIRECTED(intr)||pic_ilevel<=PIC_IRQ1||in_dpmi)
 #endif /* not USE_NEW_INT */
     {
-#if 1 /* BUG CATCHER (if 1) */
-/* outputting more then one character here will change dynamic behave such that
- * that we mesure the wrong thing.
+#if 0 /* BUG CATCHER (if 1) */
+/* outputting more then one character here will change dynamic behave such
+ * that we measure the wrong thing.
  */
 g_printf("+%d",(int)pic_ilevel);
 #endif
@@ -869,7 +877,7 @@ int pic_request(inum)
 int inum;
 {
 static char buf[81];
-  int ret=0;
+  int ret=PIC_REQ_NOP;
 
   if (pic_iinfo[inum].func == (void *)0)
     return ret; 
@@ -882,9 +890,11 @@ static char buf[81];
 #endif
     {
     if (pic_pirr&(1<<inum)){
+     ret=PIC_REQ_LOST;
      pic_print(2,"Requested irq lvl ",    inum, " lost     ");
       }
     else {
+     ret=PIC_REQ_PEND;
      pic_print(2,"Requested irq lvl ",    inum, " pending  ");
       }
     pic_pirr|=(1<<inum);
@@ -893,13 +903,13 @@ static char buf[81];
        pic_itime[inum] = pic_itime[32];
     }
     pic_ltime[inum] = pic_itime[inum];
- }
+  }
   else {
     pic_print(2,"Requested irq lvl ",    inum, " successfully");
     pic_irr|=(1<<inum);
     if(pic_itime[inum] == pic_ltime[inum]) pic_itime[inum] = pic_itime[32];
     pic_ltime[inum] = pic_itime[inum];
-    ret=1;
+    ret=PIC_REQ_OK;
   }
   if (d.request&2) {
     /* avoid going through sprintf for non-debugging */
@@ -994,10 +1004,7 @@ unsigned long pic_newirr;
 inline void pic_watch(s_time)
 hitimer_u *s_time;	/* time in us, 64-bit unsigned */
 {
-#ifndef MONOTON_MICRO_TIMING
-int timer;
-#endif
-long long t_time;
+hitimer_t t_time;
 unsigned long pic_newirr;
 
   pic_newirr=pic_wirr&~pic_irr&~pic_isr;
@@ -1011,38 +1018,11 @@ unsigned long pic_newirr;
  *  non-monoton: values are kept modulo (tick_rate*15min)
  *  monoton:     values are kept modulo 2^32 (exactly 1 hour)
  */
-#ifndef MONOTON_MICRO_TIMING
-#if 0	/* original code, see comments in timers.h */
-  t_time = (s_time->tv_sec%900)*1193047 
-           + (s_time->tv_usec*1193)/1000  /* This is usec * 1.193047 split */
-           + s_time->tv_usec/21277;        /* up to fit in 32 bit integer math */
-#endif
-  t_time = nmUStoTICK(s_time->td % 900000000);
-#else /* MONOTON_MICRO_TIMING */
   t_time = UStoTICK(s_time->td);
-#endif /* MONOTON_MICRO_TIMING */
 
   /* check for any freshly initiated timers, and sync them to s_time */
   pic_print(2,"pic_itime[1]= ",pic_itime[1]," ");
-#ifndef MONOTON_MICRO_TIMING
-  /* Now check for wrap-around, and adjust all counters if needed.  Adjustment
-     value is 900*1193047, or total counts in 15 min. */
-  if(t_time<pic_sys_time) {
-     for (timer=0;timer<33;++timer) { 
-       if(pic_itime[timer]>=pic_dos_time) {
-	 if (pic_itime[timer] != NEVER)
-          pic_itime[timer] -= 900*PIC_TICK_RATE;
-         if (pic_ltime[timer] != NEVER)
-          pic_ltime[timer] -= 900*PIC_TICK_RATE;
-          }
-       else pic_ltime[timer] = pic_itime[timer] = NEVER;
-     }
-  }
-  pic_sys_time=t_time;
-#else  /* MONOTON_MICRO_TIMING */
-  /* wraparound now is smoothly at 1h interval */
   pic_sys_time=t_time + (t_time == NEVER);
-#endif /* MONOTON_MICRO_TIMING */
   pic_print(2,"pic_sys_time set to ",pic_sys_time," ");
   pic_dos_time = pic_itime[32];
   if(pic_icount<=pic_icount_od) pic_activate();
@@ -1079,31 +1059,23 @@ int pic_pending(int ilevel)
  */
 void pic_activate()
 {
-int earliest, timer, count;
+hitimer_t earliest;
+int timer, count;
+
 /*if(pic_irr&~pic_imr) return;*/
    earliest = pic_sys_time;
    count = 0;
    for (timer=0; timer<32; ++timer) { 
-#ifndef MONOTON_MICRO_TIMING
-      if ((pic_itime[timer] < pic_sys_time) && (pic_itime[timer] != NEVER)) {
-#else /* MONOTON_MICRO_TIMING */
-      if (((int)(pic_itime[timer] - pic_sys_time) < 0) &&
-	  (pic_itime[timer] != NEVER)) {
-#endif /* MONOTON_MICRO_TIMING */
-         if( pic_itime[timer] != pic_ltime[timer]) {
-#ifndef MONOTON_MICRO_TIMING
-               if ((pic_itime[timer] < earliest) || (earliest == NEVER))
-#else /* MONOTON_MICRO_TIMING */
-               if (((int)(pic_itime[timer] - earliest) < 0) ||
-		   (earliest == NEVER))
-#endif /* MONOTON_MICRO_TIMING */
+      if ((pic_itime[timer] != NEVER) && (pic_itime[timer] < pic_sys_time)) {
+         if (pic_itime[timer] != pic_ltime[timer]) {
+               if ((earliest == NEVER) || (pic_itime[timer] < earliest))
                     earliest = pic_itime[timer];
                pic_request(timer);
                ++count;
          }
       }
    }
-   if(count) pic_print(2,"Activated ",count, "interrupts.");
+   if(count) pic_print(2,"Activated ",count, " interrupts.");
    pic_print(2,"Activate ++ dos time to ",earliest, " ");
    /*if(!pic_icount)*/ pic_dos_time = pic_itime[32] = earliest;
 }
@@ -1111,7 +1083,6 @@ int earliest, timer, count;
 /* DANG_BEGIN_FUNCTION pic_sched
  * pic_sched schedules an interrupt for activation after a designated
  * time interval.  The time measurement is in unis of 1193047/second,
- * ( or if using MONOTON_MICRO_TIMING in units of PIT_TICK_RATE/second )
  * the same rate as the pit counters.  This is convenient for timer
  * emulation, but can also be used for pacing other functions, such as
  * serial emulation, incoming keystrokes, or video updates.  Some sample 
@@ -1217,7 +1188,7 @@ void pic_reset(void)
   pic_seti(PIC_IRQ0, timer_int_engine, 0);  /* do_irq0 in pic.c */
   pic_unmaski(PIC_IRQ0);
   pic_request(PIC_IRQ0);                    /* start timer */
-  if (mice->intdrv || mice->type == MOUSE_PS2) {
+  if (mice->intdrv || mice->type == MOUSE_PS2 || mice->type == MOUSE_IMPS2) {
     pic_seti(PIC_IMOUSE, DOSEMUMouseEvents, 0);
     pic_unmaski(PIC_IMOUSE);
   }
@@ -1236,12 +1207,12 @@ void pic_reset(void)
 #endif
 }
 
-#undef set_pic0_imr(x)
-#undef set_pic1_imr(x)
-#undef get_pic0_imr()
-#undef get_pic1_imr()
-#undef get_pic0_isr()
-#undef get_pic1_isr()
-#undef get_pic0_irr()
-#undef get_pic1_irr()
+#undef set_pic0_imr
+#undef set_pic1_imr
+#undef get_pic0_imr
+#undef get_pic1_imr
+#undef get_pic0_isr
+#undef get_pic1_isr
+#undef get_pic0_irr
+#undef get_pic1_irr
 #undef NEVER
