@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -11,33 +11,17 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <sys/mman.h>
 #include <sys/times.h>
 #include <sys/time.h>
 #include <sys/stat.h>
 #include <sys/utsname.h>
 #include <errno.h>
-
-#ifdef __linux__
-#if GLIBC_VERSION_CODE >= 2000
 #include <sys/utsname.h>
-#define new_utsname utsname
-#else
-#include <linux/config.h>
-#include <linux/utsname.h>
-#endif /* GLIBC_VERSION_CODE */
-#endif /* __linux__ */
-
-#ifdef __NetBSD__
-#include <sys/utsname.h>
-#define new_utsname utsname
-#endif
 
 #include "config.h"
 #include "emu.h"
 #include "memory.h"
 #include "config.h"
-#include "kversion.h"
 #include "bios.h"
 #include "int.h"
 #include "timers.h"
@@ -63,6 +47,9 @@
 
 #include "keyb_clients.h"
 
+#include "mapping.h"
+
+
 #ifdef USING_NET
 extern void pkt_check_receive_quick(void);
 /* flag to activate use of pic by packet driver */
@@ -211,7 +198,7 @@ void hardware_setup(void)
   pic_unmaski(PIC_IRQ1);
   pic_seti(PIC_IRQ8, rtc_int8, 0);
   pic_unmaski(PIC_IRQ8);
-  if (mice->intdrv || mice->type == MOUSE_PS2) {
+  if (mice->intdrv || mice->type == MOUSE_PS2 || mice->type == MOUSE_IMPS2) {
     pic_seti(PIC_IMOUSE, DOSEMUMouseEvents, 0);
     pic_unmaski(PIC_IMOUSE);
   }
@@ -246,6 +233,8 @@ static inline void map_video_bios(void)
 {
   extern int load_file(char *name, int foffset, char *mstart, int msize);
 
+  v_printf("Mapping VBIOS = %d\n",config.mapped_bios);
+
   if (config.mapped_bios) {
     if (config.vbios_file) {
       warn("WARN: loading VBIOS %s into mem at 0x%X (0x%X bytes)\n",
@@ -282,17 +271,14 @@ static inline void map_hardware_ram(void)
 {
   int i, j;
   unsigned int addr, size;
-#ifdef __NetBSD__
-  extern int errno;
-#endif
 
+  open_mapping(MAPPING_INIT_HWRAM);
   if (!config.must_spare_hardware_ram)
     return;
   if (!can_do_root_stuff) {
     fprintf(stderr, "can't use hardware ram in low feature (non-suid root) DOSEMU\n");
     return;
   }
-  open_kmem ();
   i = 0;
   do {
     if (config.hardware_pages[i++]) {
@@ -301,33 +287,37 @@ static inline void map_hardware_ram(void)
 	i++;		/* NOTE: last byte is always ZERO */
       addr = HARDWARE_RAM_START + (j << 12);
       size = (i - j) << 12;
-      if (mmap((caddr_t) addr, (size_t) size, PROT_READ | PROT_WRITE, 
-	       MAP_SHARED | MAP_FIXED, mem_fd, addr) == (caddr_t) -1) {
+      if ((int)mmap_mapping(MAPPING_INIT_HWRAM | MAPPING_KMEM, (caddr_t) addr,
+		size, PROT_READ | PROT_WRITE, (caddr_t) addr) == -1) {
 	error("mmap error in map_hardware_ram %s\n", strerror (errno));
-	close_kmem();
 	return;
       }
       g_printf("mapped hardware ram at 0x%05x .. 0x%05x\n", addr, addr+size-1);
     }
   } while (i < sizeof (config.hardware_pages) - 1);
-  close_kmem();
 }
 
 /*
  * DANG_BEGIN_FUNCTION map_custom_bios
  *
  * description:
- *  Setup the dosemu amazing custom BIOS
+ *  Setup the dosemu amazing custom BIOS, quietly overwriting anything
+ *  was copied there before. Do not overwrite graphic fonts!
  * 
  * DANG_END_FUNCTION
  */
 static inline void map_custom_bios(void)
 {
-  extern void bios_f000(), bios_f000_end();
   u_char *ptr;
+  u_long n;
 
+  n = (u_long)bios_f000_endpart1 - (u_long)bios_f000;
   ptr = (u_char *) (BIOSSEG << 4);
-  memcpy(ptr, bios_f000, (u_long)bios_f000_end - (u_long)bios_f000);
+  memcpy(ptr, bios_f000, n);
+
+  n = (u_long)bios_f000_end - (u_long)bios_f000_part2;
+  ptr = (u_char *) (BIOSSEG << 4) + ((u_long)bios_f000_part2 - (u_long)bios_f000);
+  memcpy(ptr, bios_f000_part2, n);
 }
 
 /* 
@@ -349,7 +339,7 @@ static inline void bios_mem_setup(void)
   CONF_NFLOP(configuration, config.fdisks);
   CONF_NSER(configuration, config.num_ser);
   CONF_NLPT(configuration, config.num_lpt);
-  if ((mice->type == MOUSE_PS2) || (mice->intdrv))
+  if ((mice->type == MOUSE_PS2) || (mice->type == MOUSE_IMPS2) || (mice->intdrv))
     configuration |= CONF_MOUSE;
 
   if (config.mathco)
@@ -380,16 +370,6 @@ void memory_init(void)
   /* first_call is a truly awful hack to keep some of these procedures from
      being called twice.  It should be fixed sometime. */
 
-#if 0
-  if( first_call == 0) {
-#ifdef __linux__
-    /* make interrupt vectors read-write */
-    mprotect((void *)(BIOSSEG<<4), 0x1000, PROT_WRITE|PROT_READ|PROT_EXEC);
-    /* make memory area 0xf8000 to 0xfffff read write */
-    mprotect((void *)(ROMBIOSSEG<<4), 0x8000, PROT_WRITE|PROT_READ|PROT_EXEC);
-#endif
-  }
-#endif
   map_video_bios();            /* map the video bios */
   if (first_call)
     map_hardware_ram();        /* map the direct hardware ram */
@@ -441,12 +421,6 @@ void memory_init(void)
     shared_memory_init();
   }
   first_call = 0;
-#if 0
-  /* make interrupts vector read-only */
-  mprotect((void *)(BIOSSEG<<4), 0x1000, PROT_READ|PROT_EXEC);
-  /* make memory area 0xf8000 to 0xfffff read only */
-  mprotect((void *)(ROMBIOSSEG<<4), 0x8000, PROT_READ|PROT_EXEC);
-#endif
 }
 
 /* 
@@ -499,11 +473,11 @@ void device_init(void)
 void low_mem_init(void)
 {
   char *result;
+
+  open_mapping(MAPPING_INIT_LOWRAM);
   g_printf ("DOS memory area being mapped in\n");
-  result = mmap (NULL, 0x100000,
-                 PROT_EXEC | PROT_READ | PROT_WRITE,
-                 MAP_FIXED | MAP_PRIVATE | MAP_ANON,
-                 -1, 0);
+  result = mmap_mapping(MAPPING_INIT_LOWRAM | MAPPING_SCRATCH, 0,
+  		0x100000, PROT_EXEC | PROT_READ | PROT_WRITE, 0);
   if (result != NULL)
     {
       perror ("anonymous mmap");
@@ -522,7 +496,7 @@ void low_mem_init(void)
  * DANG_END_FUNCTION
  */
 void version_init(void) {
-  struct new_utsname unames;
+  struct utsname unames;
   char version[80];
 
   uname((struct utsname *)&unames);
@@ -552,14 +526,8 @@ void version_init(void) {
 
 void print_version(void)
 {
-  struct new_utsname unames;
+  struct utsname unames;
     
   uname((struct utsname *)&unames);
   warn("DOSEMU-%s is coming up on %s version %s\n", VERSTR, unames.sysname, unames.release);
-#ifdef __linux__
-  warn("Built for %d\n", LX_KERNEL_VERSION);
-#endif
-#ifdef __NetBSD__
-  warn("Built for %d\n", NETBSD_VERSION);
-#endif
 }
