@@ -1,14 +1,15 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <unistd.h>
-#include <linux/kd.h>
-#include <linux/keyboard.h>
+#include <sys/kd.h>
+#include "Linux/keyboard.h"
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
