@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -15,15 +15,13 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <malloc.h>
+#include <ctype.h>
 #include <string.h>
 #include <sys/ioctl.h>
 #ifdef __linux__
-#include <linux/hdreg.h>
-#include <linux/fd.h>
-#include <linux/fs.h>
-#endif
-#ifdef __NetBSD__
-#include "netbsd_disk.h"
+  #include "Linux/hdreg.h"
+  #include "Linux/fd.h"
+  #include "Linux/fs.h"
 #endif
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -31,6 +29,18 @@
 #include "emu.h"
 #include "disks.h"
 #include "priv.h"
+#include "dosemu_select.h"
+#include "int.h"
+#include "fatfs.h"
+#include "utilities.h"
+
+#ifdef NEED_LLSEEK_PROTOTYPE
+  /* well, if we don't have llseek prototype,
+   * we most likely won't have __loff_t too, hence using long long
+   */
+  extern long long libless_llseek(int fd, long long offset, int origin);
+  #define llseek libless_llseek
+#endif
 
 static int disks_initiated = 0;
 
@@ -68,10 +78,34 @@ static struct disk_fptr disk_fptrs[NUM_DTYPES] =
   {image_auto, image_setup},
   {hdisk_auto, hdisk_setup},
   {floppy_auto, floppy_setup},
-  {partition_auto, partition_setup}
+  {partition_auto, partition_setup},
+  {dir_auto, dir_setup}
 };
 
 
+static void dump_disk_blks(unsigned char *tb, int count, int ssiz)
+{
+  static unsigned char buf[80], cbuf[20];
+  int a,i,j;
+  unsigned char *p,*q;
+
+  q=tb; a=0;
+  while (count--) {
+    for (i=0; i < (ssiz>>4); i++) {
+      p=buf;
+      p+=sprintf(p,"%04x: ",a);
+      for (j=0; j<16; j++) {
+        p+=sprintf(p,"%02x ",*q);
+        cbuf[j] = (isprint(*q)? *q:'.');
+        q++;
+      }
+      cbuf[16]=0; a+=16;
+      d_printf("%s  %s\n",buf,cbuf);
+    }
+    d_printf("\n");
+  }
+}
+
 /* read_sectors
  *
  * okay, here's the purpose of this: to handle reads orthogonally across
@@ -81,7 +115,7 @@ static struct disk_fptr disk_fptrs[NUM_DTYPES] =
  *
  *   |   adds up to 1 sector (512 bytes) of Master Boot Record (MBR)
  *       THIS IS KEPT IN MEMORY, and is the file 
- *       /var/lib/dosemu/partition.<partition>.
+ *       <DOSEMU_LIB_DIR>/partition.<partition>.
  *
  *   .   is any number of sectors before the desired partition start address.
  *       THESE SECTORS ARE NOT EMULATED, BUT MERELY SKIPPED
@@ -106,93 +140,91 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
 
   /* XXX - header hack. dp->header can be negative for type PARTITION */
   pos = DISK_OFFSET(dp, head, sector, track) + dp->header;
-  d_printf("DISK: Trying to read %ld bytes at T/S/H %ld/%ld/%ld",
-	   count,track,sector,head);
+  d_printf("DISK: %s: Trying to read %ld sectors at T/S/H %ld/%ld/%ld",
+	   dp->dev_name,count,track,sector,head);
 #ifdef __linux__
-  d_printf(" at pos %Ld\n", pos);
+  d_printf("%+Ld at pos %Ld\n", dp->header, pos);
 #else
   d_printf(" at pos %ld\n", pos);
 #endif
 
-
   /* reads beginning before that actual disk/file */
-  if (pos < 0) {
+  if (pos < 0 && count > 0) {
     int readsize = count * SECTOR_SIZE;
     int mbroff = -dp->header + pos;
-    int mbrcount, mbrinc;
+    int mbrread = 0;
 
-    if (dp->type != PARTITION) {
+    if(!(dp->type == PARTITION || dp->type == DIR_TYPE)) {
       error("negative offset on non-partition disk type\n");
       return -DERR_NOTFOUND;
     }
 
-    if (readsize >= -pos) {
-      mbrcount = -pos;
-      mbrinc = -pos;
-    }
-    else {
-      mbrcount = readsize;
-      mbrinc = readsize;
+    readsize = readsize > -pos ? -pos : readsize;
+
+    /* 
+     * this will end up pretending to read the requested data, but will
+     * only read as much as can be read from the fake MBR in memory
+     * and zero the rest
+     */
+
+    /* copy the MBR... */
+    if(mbroff < dp->part_info.mbr_size) {
+      mbrread = dp->part_info.mbr_size - mbroff;
+      mbrread = mbrread > readsize ? readsize : mbrread;
+      memcpy(buffer, dp->part_info.mbr + mbroff, mbrread);
+      d_printf("read 0x%lx bytes from MBR, ofs = 0x%lx (0x%lx bytes left)\n",
+        (unsigned long) mbrread, (unsigned long) mbroff, (unsigned long) (readsize - mbrread)
+      );
     }
 
-    /* this should take a third variable, mbr_begin, into account, but
-       * the MBR always begins at offset 0
-       * this will end up pretending to read the requested data, but will
-       * only read as much as can be read from the fake MBR in memory,
-       * simply skipping the rest.  Should I zero the "unread" parts of
-       * the dest. buffer?  No matter, I guess.
-       */
-    if ((mbroff + mbrcount) > dp->part_info.mbr_size) {
-      error("%s writing in the forbidden fake partition zone!\n",
-	    dp->dev_name);
-      mbrcount = dp->part_info.mbr_size - mbroff;
+    /* ... and zero the rest */
+    if(readsize > mbrread) {
+      memset(buffer + mbrread, 0, readsize - mbrread);
+      d_printf("emulated reading 0x%lx bytes, ofs = 0x%lx\n",
+        (unsigned long) (readsize - mbrread), (unsigned long) (mbroff + mbrread)
+      );
     }
 
-#ifdef __linux__
-    d_printf("WARNING: %Ld",pos);
-#else
-    d_printf("WARNING: %ld",pos);
-#endif
-    d_printf(" (0x%lx) read for %d below %s, "
-	     "h:%ld s:%ld t:%ld\n",
-	     (unsigned long) -pos, readsize,
-	     dp->dev_name, head, sector, track);
-
-    memcpy(buffer, dp->part_info.mbr + mbroff, mbrcount);
-
-    /* even though we may not have actually read mbrinc bytes (i.e. the
-       * read spanned non-emulated empty space), we increment this much
-       * so that we can pretend that the whole read worked and get on with
-       * the rest.
-       */
-    buffer += mbrinc;
-    pos += mbrinc;
-    already = mbrinc;
-
-    /* simulate a read() of count*SECTOR_SIZE bytes */
-    if (readsize == mbrinc) {
-      d_printf("   got entire read done from memory. off:%d, count:%d\n",
-	       mbroff, mbrcount);
-
+    if(readsize == count * SECTOR_SIZE) {
+      d_printf("   got entire read done from memory. off:%d, count:%d\n", mbroff, readsize);
       return readsize;
     }
+
+    buffer += readsize;
+    pos += readsize;
+    already += readsize;
   }
 
+  if(dp->type == DIR_TYPE && dp->fatfs) {
+    /* this should not never happen */
+    if(pos % SECTOR_SIZE || already % SECTOR_SIZE) {
+      error("illegal read offset for %s\n", dp->dev_name);
+      return -DERR_NOTFOUND;
+    }
+    tmpread = fatfs_read(dp->fatfs, buffer, pos / SECTOR_SIZE, count - already / SECTOR_SIZE);
+    if(tmpread == -1) return -DERR_NOTFOUND;
+    if(tmpread == -2) return -DERR_ECCERR;
+    tmpread *= SECTOR_SIZE;
+  }
+  else {
 #ifdef __linux__
-  if (pos != llseek(dp->fdesc, pos, SEEK_SET)) {
+    if(pos != llseek(dp->fdesc, pos, SEEK_SET)) {
 #else
-  if (pos != lseek(dp->fdesc, pos, SEEK_SET)) {
+    if(pos != lseek(dp->fdesc, pos, SEEK_SET)) {
 #endif
-    error("Sector not found in read_sector, error = %s!\n",
-	  strerror(errno));
-    return -DERR_NOTFOUND;
+      error("Sector not found in read_sector, error = %s!\n", strerror(errno));
+      return -DERR_NOTFOUND;
+    }
+    tmpread = RPT_SYSCALL(read(dp->fdesc, buffer, count * SECTOR_SIZE - already));
   }
 
-  tmpread = RPT_SYSCALL(read(dp->fdesc, buffer, count * SECTOR_SIZE));
-  if (tmpread != -1)
-    return (tmpread + already);
-  else
+  if(tmpread != -1) {
+    if(d.disk > 8) dump_disk_blks(buffer, count - already / SECTOR_SIZE, SECTOR_SIZE);
+    return tmpread + already;
+  }
+  else {
     return -DERR_ECCERR;
+  }
 }
 
 int
@@ -200,7 +232,7 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
 	      long track, long count)
 {
   loff_t  pos;
-  int tmpwrite;
+  long tmpwrite, already = 0;
 
   if (dp->rdonly) {
     d_printf("ERROR: write to readonly disk %s\n", dp->dev_name);
@@ -214,33 +246,60 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
 
   /* XXX - header hack */
   pos = DISK_OFFSET(dp, head, sector, track) + dp->header;
-  d_printf("DISK: Trying to write %ld bytes T/S/H %ld/%ld/%ld",
-	   count,track,sector,head);
+  d_printf("DISK: %s: Trying to write %ld sectors T/S/H %ld/%ld/%ld",
+	   dp->dev_name,count,track,sector,head);
 #ifdef __linux__
   d_printf(" at pos %Ld\n", pos);
 #else
   d_printf(" at pos %ld\n", pos);
 #endif
 
-  /* if dp->type is PARTITION, we currently don't allow writing to an area
-   * not within the actual partition (i.e. somewhere else on the faked
-   * disk). I could avoid this by somehow relocating the partition to
-   * the beginning of the disk, but I confess I'm not sure how I would
-   * do that.
-   *
-   * Also, I could allow writing to the fake MBR, but I'm not sure that's
-   * a good idea.
+  /*
+   * writes outside the partition (before the actual disk/file) are ignored
    */
-#ifdef __linux__
-  if (pos != llseek(dp->fdesc, pos, SEEK_SET)) {
-#else
-  if (pos != lseek(dp->fdesc, pos, SEEK_SET)) {
-#endif
-    error("Sector not found in write_sector!\n");
-    return -DERR_NOTFOUND;
+  if (pos < 0 && count > 0) {
+    long writesize = count * SECTOR_SIZE;
+
+    if(!(dp->type == PARTITION || dp->type == DIR_TYPE)) {
+      error("negative offset on non-partition disk type\n");
+      return -DERR_NOTFOUND;
+    }
+
+    writesize = writesize > -pos ? -pos : writesize;
+
+    d_printf("emulated writing 0x%lx bytes, ofs = 0x%lx (0x%lx left)\n",
+      writesize, (long) (pos - dp->header), count * SECTOR_SIZE - writesize
+    );
+
+    if(writesize == count * SECTOR_SIZE) return writesize;
+
+    buffer += writesize;
+    pos += writesize;
+    already += writesize;
   }
 
-  tmpwrite = RPT_SYSCALL(write(dp->fdesc, buffer, count * SECTOR_SIZE));
+  if(dp->type == DIR_TYPE && dp->fatfs) {
+    /* this should not never happen */
+    if(pos % SECTOR_SIZE || already % SECTOR_SIZE) {
+      error("illegal write offset for %s\n", dp->dev_name);
+      return -DERR_NOTFOUND;
+    }
+    tmpwrite = fatfs_write(dp->fatfs, buffer, pos / SECTOR_SIZE, count - already / SECTOR_SIZE);
+    if(tmpwrite == -1) return -DERR_NOTFOUND;
+    if(tmpwrite == -1) return -DERR_WRITEFLT;
+    tmpwrite *= SECTOR_SIZE;
+  }
+  else {
+#ifdef __linux__
+    if(pos != llseek(dp->fdesc, pos, SEEK_SET)) {
+#else
+    if(pos != lseek(dp->fdesc, pos, SEEK_SET)) {
+#endif
+      error("Sector not found in write_sector!\n");
+      return -DERR_NOTFOUND;
+    }
+    tmpwrite = RPT_SYSCALL(write(dp->fdesc, buffer, count * SECTOR_SIZE - already));
+  }
 
   /* this should make floppies a little safer...I would as soon use the
    * O_SYNC flag, but we don't have it.  fsync() should be in the kernel
@@ -251,7 +310,7 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
 
   FLUSHDISK(dp);
 
-  return tmpwrite;
+  return tmpwrite + already;
 }
 
 void
@@ -310,24 +369,6 @@ image_auto(struct disk *dp)
 void
 hdisk_auto(struct disk *dp)
 {
-#ifdef __NetBSD__
-  struct disklabel label;
-  struct stat stb;
-
-  if (ioctl(dp->fdesc, DIOCGDINFO, &label) != 0 ||
-      fstat(dp->fdesc, &stb) != 0) {
-      error("can't fstat %s: %s\n", dp->dev_name,
-	    strerror(errno));
-      leavedos(21);
-  } else {
-    dp->sectors = label.d_nsectors;
-    dp->heads = label.d_ntracks;
-    dp->tracks = label.d_ncylinders;
-    dp->start = label.d_partitions[DISKPART(stb.st_rdev)].p_offset;
-    d_printf("HDISK auto_info disk %s; h=%d, s=%d, t=%d, start=%d\n",
-	     dp->dev_name, dp->heads, dp->sectors, dp->tracks, dp->start);
-  }
-#endif
 #ifdef __linux__
   struct hd_geometry geo;
 
@@ -347,6 +388,90 @@ hdisk_auto(struct disk *dp)
 #endif
 }
 
+void dir_auto(struct disk *dp)
+{
+  /*
+   * We emulate an entire disk with 1 partition starting at t/h/s 0/1/1.
+   * You are free to change the geometry (e.g. to change the partition size).
+   */
+  dp->sectors = 63;
+  dp->heads = 255;
+  dp->tracks = 50;
+
+  dp->start = dp->sectors;
+
+  d_printf(
+    "DIR auto_info disk %s; h=%d, s=%d, t=%d, start=%ld\n",
+    dp->dev_name, dp->heads, dp->sectors, dp->tracks, dp->start
+  );
+}
+
+void dir_setup(struct disk *dp)
+{
+  unsigned char *mbr;
+  struct partition *pi = &dp->part_info;
+  int i = strlen(dp->dev_name);
+
+  while(--i >= 0) if(dp->dev_name[i] == '/') dp->dev_name[i] = 0; else break;
+
+  d_printf("partition setup for directory %s\n", dp->dev_name);
+
+  pi->beg_head = 1;
+  pi->beg_sec = 1;
+  pi->beg_cyl = 0;
+  pi->end_head = dp->heads - 1;
+  pi->end_sec = dp->sectors;
+  pi->end_cyl = dp->tracks - 1;
+  pi->pre_secs = dp->sectors;
+  pi->num_secs = (dp->tracks * dp->heads - 1) * dp->sectors;
+
+  dp->header = -(SECTOR_SIZE * (loff_t) (pi->pre_secs));
+
+  pi->mbr_size = SECTOR_SIZE;
+  pi->mbr = malloc(pi->mbr_size);
+  mbr = pi->mbr;
+
+  memset(mbr, 0, SECTOR_SIZE);
+  /*
+   * mov ax,0fffeh
+   * int 0e6h
+   */
+  mbr[0x00] = 0xb8;
+  mbr[0x01] = 0xfe;
+  mbr[0x02] = 0xff;
+  mbr[0x03] = 0xcd;
+  mbr[0x04] = 0xe6;
+  mbr[PART_INFO_START + 0x00] = PART_BOOT;
+  mbr[PART_INFO_START + 0x01] = pi->beg_head;
+  mbr[PART_INFO_START + 0x02] = pi->beg_sec + ((pi->beg_cyl & 0x300) >> 2);
+  mbr[PART_INFO_START + 0x03] = pi->beg_cyl;
+  mbr[PART_INFO_START + 0x04] = pi->num_secs < 1 << 16 ? 0x04 : 0x06;
+  mbr[PART_INFO_START + 0x05] = pi->end_head;
+  mbr[PART_INFO_START + 0x06] = pi->end_sec + ((pi->end_cyl & 0x300) >> 2);
+  mbr[PART_INFO_START + 0x07] = pi->end_cyl;
+  *(unsigned *) (mbr + PART_INFO_START + 0x08) = pi->pre_secs;
+  *(unsigned *) (mbr + PART_INFO_START + 0x0c) = pi->num_secs;
+  mbr[SECTOR_SIZE - 2] = 0x55;
+  mbr[SECTOR_SIZE - 1] = 0xaa;
+
+  d_printf("partition table entry for device %s is:\n", dp->dev_name);
+  d_printf(
+    "beg head %d, sec %d, cyl %d = end head %d, sec %d, cyl %d\n",
+    pi->beg_head, pi->beg_sec, pi->beg_cyl,
+    pi->end_head, pi->end_sec, pi->end_cyl
+  );
+  d_printf(
+    "pre_secs %ld, num_secs %ld = %lx, -dp->header %ld = 0x%lx\n",
+    pi->pre_secs, pi->num_secs, pi->num_secs,
+    (long) -dp->header, (unsigned long) -dp->header
+  );
+
+  dp->fatfs = NULL;
+  fatfs_init(dp);
+
+  if(dp->fatfs && !redir_state) set_int21_revectored(redir_state = 1);
+}
+
 /* XXX - relies upon a file of SECTOR_SIZE in PARTITION_PATH that which
  *       contains the MBR (first sector) of the drive (i.e. /dev/hda).
  *       only works for /dev/hda1 right now, and only allows one
@@ -375,20 +500,16 @@ partition_setup(struct disk *dp)
 
   d_printf("PARTITION SETUP for %s\n", dp->dev_name);
 
-#ifdef __NetBSD__
-  hd_name = strdup(dp->dev_name);
-  hd_name[strlen(hd_name)-1] = 'd';	/* i.e.  /dev/rwd0g -> /dev/rwd0d */
-  PNUM = 5;				/* XXX */
-#endif
 #ifdef __linux__
   hd_name = strdup(dp->dev_name);
   hd_name[8] = '\0';			/* i.e.  /dev/hda6 -> /dev/hda */
 #endif
 
   enter_priv_on();
-  part_fd = DOS_SYSCALL(open(hd_name, O_RDONLY));
+  part_fd = SILENT_DOS_SYSCALL(open(hd_name, O_RDONLY));
   leave_priv_setting();
   if (part_fd == -1) {
+    if (dp->removeable) return;
     error("opening device %s to read MBR for PARTITION %s\n",
 	  hd_name, dp->dev_name);
     leavedos(22);
@@ -473,18 +594,6 @@ set_part_ent(struct disk *dp, char *tmp_mbr)
   long	end;		/* last sector number offset		*/
   char	*p;		/* ptr to part table entry to create	*/
 
-#ifdef __NetBSD__
-  struct disklabel label;
-  struct stat stb;
-
-  if (ioctl(dp->fdesc, DIOCGDINFO, &label) != 0 ||
-      fstat(dp->fdesc, &stb) != 0) {
-      error("can't fstat %s: %s\n", dp->dev_name,
-	    strerror(errno));
-      leavedos(21);
-  } else
-      length = label.d_partitions[DISKPART(stb.st_rdev)].p_size;
-#endif
 #ifdef __linux__
   if (ioctl(dp->fdesc, BLKGETSIZE, &length)) {
     error("calling ioctl BLKGETSIZE for PARTITION %s\n", dp->dev_name);
@@ -525,6 +634,9 @@ d_nullf(struct disk *dp)
   d_printf("NULLF for %s\n", dp->dev_name);
 }
 
+
+unsigned char ATAPI_buf0[512] = { 0 };
+
 inline void
 disk_close(void)
 {
@@ -533,84 +645,13 @@ disk_close(void)
   if (!disks_initiated) return;  /* just to be safe */
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
     if (dp->removeable && dp->fdesc >= 0) {
-      d_printf("DISK: Closing a disk\n");
+      d_printf("DISK: %s: Closing a disk\n",dp->dev_name);
       (void) close(dp->fdesc);
       dp->fdesc = -1;
     }
   }
 }
 
-#ifdef __NetBSD__
-void
-disk_open(struct disk *dp)
-{
-  PRIV_SAVE_AREA
-  struct fd_type fl;
-
-  if (dp == NULL || dp->fdesc >= 0)
-    return;
-    
-  enter_priv_on();
-  dp->fdesc = DOS_SYSCALL(open(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
-  leave_priv_setting();
-  if (dp->fdesc < 0) 
-    if (errno == EROFS || errno == EACCES) {
-      enter_priv_on();
-      dp->fdesc = DOS_SYSCALL(open(dp->dev_name, O_RDONLY, 0));
-      leave_priv_setting();
-      if (dp->fdesc < 0) {
-        d_printf("ERROR: (disk) can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
-        /* In case we DO get more clever, we want to share that code */
-        goto fail;
-      } else {
-        dp->rdonly = 1;
-        d_printf("(disk) can't open %s for read/write. Readonly did work though\n", dp->dev_name);
-      }
-    } else {
-      d_printf("ERROR: (disk) can't open %s: %s\n", dp->dev_name, strerror(errno));
-    fail:
-#if 0
-      /* We really should be more clever here */
-      fatalerr = 5;
-#endif
-      return;
-    }
-  else dp->rdonly = dp->wantrdonly;
-
-  if (ioctl(dp->fdesc, FD_GTYPE, &fl) == -1) {
-      int err = errno;
-      d_printf("ERROR: floppy gettype: %s\n", strerror(err));
-    if (err == ENODEV || err == EIO) {	/* no disk available */
-      dp->sectors = 0;
-      dp->heads = 0;
-      dp->tracks = 0;
-      return;
-    }
-    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, strerror(err));
-    fatalerr = 5;
-    return;
-  }
-  d_printf("FLOPPY %s h=%d, s=%d, t=%d\n",
-	   dp->dev_name, fl.heads, fl.sectrac, fl.tracks);
-  dp->sectors = fl.sectrac;
-  dp->heads = fl.heads;
-  dp->tracks = fl.tracks;
-  /* XXX turn media change msgs off? */
-  /*  DOS_SYSCALL(ioctl(dp->fdesc, FDMSGOFF, 0));*/
-}
-#endif
-
-static void sigalarm_onoff(int on)
-{
-  static struct itimerval itv_old;
-  static struct itimerval itv;
-  if (on) setitimer(TIMER_TIME, &itv_old, NULL);
-  else {
-    itv.it_interval.tv_sec = itv.it_interval.tv_usec = 0;
-    itv.it_value = itv.it_interval;
-    setitimer(TIMER_TIME, &itv, &itv_old);
-  }
-}
 
 #ifdef __linux__
 void
@@ -623,10 +664,15 @@ disk_open(struct disk *dp)
     return;
     
   enter_priv_on();
-  dp->fdesc = SILENT_DOS_SYSCALL(open(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
+  dp->fdesc = SILENT_DOS_SYSCALL(open(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
   leave_priv_setting();
 
-  if (dp->fdesc < 0) {
+  /* FIXME:
+   * Why the hell was the below handling restricted to non-removeable disks?
+   * This made opening writeprotected floppies impossible :-(
+   *                                                  -- Hans, 990112
+   */
+  if ( /*!dp->removeable &&*/ (dp->fdesc < 0)) {
     if (errno == EROFS || errno == ENODEV) {
       enter_priv_on();
       dp->fdesc = DOS_SYSCALL(open(dp->dev_name, O_RDONLY, 0));
@@ -660,15 +706,33 @@ disk_open(struct disk *dp)
    * To overcome this problem we temporary switch off the timer
    * during the ioctl. (well, not what we really like)
    * ( 19 May 1996, Hans Lermen ) */
-  int res;
+  int res=0;
+  if (dp->default_cmos == ATAPI_FLOPPY) {
+	unsigned long tns;
+	if (ATAPI_buf0[0] || (read_sectors(dp, ATAPI_buf0, 0, 0, 0, 1) > 0)) {
+	      fl.sect = *((unsigned char *)&ATAPI_buf0[0x18]);
+	      fl.head = *((unsigned char *)&ATAPI_buf0[0x1a]);
+	      tns = *((unsigned short *)&ATAPI_buf0[0x13]);
+	      if (tns==0) tns = *((unsigned long *)&ATAPI_buf0[0x20]);
+	      fl.track = tns/(fl.sect*fl.head);
+	}
+	else {	/* no disk available */
+	      dp->sectors = 0;
+	      dp->heads = 0;
+	      dp->tracks = 0;
+	      return;
+	}
+  }
+  else {
   sigalarm_onoff(0);
   res = ioctl(dp->fdesc, FDGETPRM, &fl);
   sigalarm_onoff(1);
+  }
   if (res == -1) {
 #else
   if (ioctl(dp->fdesc, FDGETPRM, &fl) == -1) {
 #endif
-    if (errno == ENODEV) {	/* no disk available */
+    if ((dp->fdesc == -1) || (errno == ENODEV)) {	/* no disk available */
       dp->sectors = 0;
       dp->heads = 0;
       dp->tracks = 0;
@@ -683,7 +747,8 @@ disk_open(struct disk *dp)
   dp->sectors = fl.sect;
   dp->heads = fl.head;
   dp->tracks = fl.track;
-  DOS_SYSCALL(ioctl(dp->fdesc, FDMSGOFF, 0));
+  if (dp->default_cmos != ATAPI_FLOPPY)
+    DOS_SYSCALL(ioctl(dp->fdesc, FDMSGOFF, 0));
 }
 #endif
 
@@ -700,6 +765,7 @@ disk_close_all(void)
     d_printf("BOOTDISK Closing\n");
   }
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
+    ATAPI_buf0[0] = 0;
     if (dp->fdesc >= 0) {
       d_printf("Floppy disk Closing %x\n", dp->fdesc);
       (void) close(dp->fdesc);
@@ -707,6 +773,7 @@ disk_close_all(void)
     }
   }
   for (dp = hdisktab; dp < &hdisktab[HDISKS]; dp++) {
+    if(dp->type == DIR_TYPE) fatfs_done(dp);
     if (dp->fdesc >= 0) {
       d_printf("Hard disk Closing %x\n", dp->fdesc);
       (void) close(dp->fdesc);
@@ -756,34 +823,32 @@ disk_init(void)
           d_printf("(disk) can't open bootdisk %s for read/write. Readonly did work though\n", bootdisk.dev_name);
         }
       } else {
-        error("can't open bootdisk %s: %sn", dp->dev_name, strerror(errno));
+        error("can't open bootdisk %s: %s\n", dp->dev_name, strerror(errno));
         leavedos(23);
       }
     }
     else bootdisk.rdonly = bootdisk.wantrdonly;
     bootdisk.removeable = 0;
   }
+
+  /*
+   * Open floppy disks
+   */
+  ATAPI_buf0[0] = 0;
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
     if (stat(dp->dev_name, &stbuf) < 0) {
       error("can't stat %s\n", dp->dev_name);
       leavedos(24);
     }
     if (S_ISBLK(stbuf.st_mode))
-      d_printf("ISBLK\n");
+      d_printf("ISBLK ");
     if (S_ISCHR(stbuf.st_mode))
-      d_printf("ISCHR\n");
-    d_printf("dev : %x\n", stbuf.st_rdev);
-#ifdef __NetBSD__
-    if ((S_ISBLK(stbuf.st_mode) && major(stbuf.st_rdev) == 0x2) ||
-	(S_ISCHR(stbuf.st_mode) && major(stbuf.st_rdev) == 0x9)) {
-      d_printf("DISK %s removeable\n", dp->dev_name);
-      dp->removeable = 1;
-      dp->fdesc = -1;
-      continue;
-    }
-#endif
+      d_printf("ISCHR ");
+    d_printf("dev %s: %#x\n", dp->dev_name, (unsigned) stbuf.st_rdev);
 #ifdef __linux__
-    if (S_ISBLK(stbuf.st_mode) && (stbuf.st_rdev & 0xff00) == 0x200) {
+    if (S_ISBLK(stbuf.st_mode) && 
+    (((stbuf.st_rdev & 0xff00)==0x200) || (dp->default_cmos==ATAPI_FLOPPY))
+    ) {
       d_printf("DISK %s removeable\n", dp->dev_name);
       dp->removeable = 1;
       dp->fdesc = -1;
@@ -807,11 +872,16 @@ disk_init(void)
         }
       } else {
         error("can't open %s: %s\n", dp->dev_name, strerror(errno));
+        /* 'leavedos' here hangs the machine */
         leavedos(25);
       }
     }
     else dp->rdonly = dp->wantrdonly;
   }
+
+  /*
+   * Open hard disks
+   */
   for (dp = hdisktab; dp < &hdisktab[HDISKS]; dp++) {
     if(dp->type == IMAGE)  {
 	if (dp->dexeflags & DISK_DEXE_RDWR) {
@@ -824,7 +894,7 @@ disk_init(void)
 	}
     }
     else enter_priv_on();
-    dp->fdesc = open(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
+    dp->fdesc = open(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
     leave_priv_setting();
     if (dp->fdesc < 0) {
       if (errno == EROFS || errno == EACCES) {
@@ -907,11 +977,11 @@ int
 checkdp(struct disk *disk)
 {
   if (disk == NULL) {
-    error("DISK: null dp\n");
+    d_printf("DISK: null dp\n");
     return 1;
   }
   else if (disk->fdesc == -1) {
-    error("DISK: closed disk\n");
+    d_printf("DISK: closed disk\n");
     return 1;
   }
   else
@@ -981,14 +1051,10 @@ int13(u_char i)
       track |= (HI(dx) & 0xc0) << 4;
     buffer = SEG_ADR((char *), es, bx);
     number = LO(ax);
-#if 0
-    d_printf("DISK %d read [h:%d,s:%d,t:%d](%d)->%p\n",
-	     disk, head, sect, track, number, (void *) buffer);
-#endif
 
     if (checkdp_val || head >= dp->heads ||
 	sect >= dp->sectors || track >= dp->tracks) {
-      error("Sector not found 1!\n");
+      d_printf("Sector not found 1!\n");
       d_printf("DISK %d read [h:%d,s:%d,t:%d](%d)->%p\n",
 	       disk, head, sect, track, number, (void *) buffer);
       if (dp) {
@@ -1052,7 +1118,7 @@ int13(u_char i)
     }
 
     if (dp->rdonly) {
-      error("write protect!\n");
+      W_printf("write protect!\n");
       show_regs(__FILE__, __LINE__);
       if (dp->removeable)
 	HI(ax) = DERR_WP;
@@ -1063,7 +1129,7 @@ int13(u_char i)
     }
 
     if (dp->rdonly)
-      error("CONTINUED!!!!!\n");
+      W_printf("CONTINUED!!!!!\n");
     res = write_sectors(dp, buffer, head, sect, track, number);
 
     if (res < 0) {
@@ -1137,33 +1203,32 @@ int13(u_char i)
     break;
 
   case 8:			/* get disk drive parameters */
-    d_printf("disk get parameters %d\n", disk);
+    d_printf("disk get parameters %#x\n", disk);
 
     if (dp != NULL) {
-      /* get CMOS type */
-      /* LO(bx) = 3; */
-      switch (dp->sectors) {
+      /* get CMOS type for floppies*/
+      if (disk < 0x80) switch (dp->sectors) {
       case 9:
-	LO(bx) = 3;
+	LO(bx) = THREE_INCH_720KFLOP;
 	break;
       case 15:
-	LO(bx) = 2;
+	LO(bx) = FIVE_INCH_FLOPPY;
 	break;
       case 18:
-	LO(bx) = 4;
+	LO(bx) = THREE_INCH_FLOPPY;
 	break;
       case 0:
 	LO(bx) = dp->default_cmos;
 	dp->tracks = 80;
 	dp->heads = 2;
-	if (LO(bx) == 4)
-	  dp->sectors = 18;
-	else
+	if (LO(bx) == FIVE_INCH_FLOPPY)
 	  dp->sectors = 15;
+	else
+	  dp->sectors = 18;
 	d_printf("auto type defaulted to CMOS %d, sectors: %d\n", LO(bx), dp->sectors);
 	break;
       default:
-	LO(bx) = 4;
+	LO(bx) = THREE_INCH_FLOPPY;
 	d_printf("type det. failed. num_tracks is: %d\n", dp->tracks);
 	break;
       }
@@ -1173,7 +1238,7 @@ int13(u_char i)
       HI(cx) = (dp->tracks - 1) & 0xff;
 
       LO(dx) = (disk < 0x80) ? FDISKS : HDISKS;
-      LO(cx) = dp->sectors | (((dp->tracks -1) & 0x300) >> 2);
+      LO(cx) = (dp->sectors & 0x3f) | (((dp->tracks -1) & 0x300) >> 2);
       LO(ax) = 0;
       HI(ax) = DERR_NOERR;
       REG(eflags) &= ~CF;	/* no error */
@@ -1243,7 +1308,7 @@ int13(u_char i)
     /* end of Alan's additions */
 
   case 0x15:			/* Get type */
-    d_printf("disk gettype %d\n", disk);
+    d_printf("disk gettype %#x\n", disk);
     if (dp != NULL && disk >= 0x80) {
       if (dp->removeable) {
 	HI(ax) = 1;		/* floppy disk, change detect (1=no, 2=yes) */
@@ -1504,7 +1569,7 @@ int13(u_char i)
     NOCARRY;
     break;
   default:
-    error("disk error, unknown command: int13, ax=0x%x\n",
+    d_printf("disk error, unknown command: int13, ax=0x%x\n",
 	  LWORD(eax));
     show_regs(__FILE__, __LINE__);
     CARRY;
@@ -1527,4 +1592,3 @@ floppy_tick(void)
     secs = 0;
   }
 }
-
