@@ -1,6 +1,13 @@
+/* 
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * 
  * DANG_BEGIN_MODULE
+ * REMARK
  *
  * This file contains a simple system for passing information through to
  * DOSEMU from the Linux side. It does not allow dynamic message passing,
@@ -15,6 +22,7 @@
  * string also contains the name of the environment variable to interrogate.
  * (The string is overwritten with the reply).
  *
+ * /REMARK
  * DANG_END_MODULE
  *
  * DANG_BEGIN_CHANGELOG
@@ -46,6 +54,7 @@
 #include "pic.h"
 #include "int.h"
 #include "vc.h"
+#include "dosemu_select.h"
 
 #ifndef max
 #define max(a,b)       ((a)>(b)? (a):(b))
@@ -149,9 +158,9 @@ static int fork_debug(void)
  * 2/9/1995, Erik Mouw (j.a.k.mouw@et.tudelft.nl):
  *  - initial version
  * 3/10/1995, Erik Mouw (j.a.k.mouw@et.tudelft.nl):
- *  - fixed securety hole
+ *  - fixed security hole
  *  - stderr output also to dosemu screen
- * 2/27/1997, Alberto Vignani (vignani@torino.alpcom.it):
+ * 2/27/1997, Alberto Vignani (vignani@mail.tin.it):
  *  - make it work (tested with 'ls -lR /' and 'updatedb &')
  *
  * DANG_BEGIN_FUNCTION run_unix_command
@@ -334,3 +343,40 @@ void run_unix_command(char *buffer)
             WEXITSTATUS(status));
     }
 }
+
+
+/*
+ * This runs system() in a forked environment with all priviledges off
+ * while scheduling vital DOSEMU functionality in the background.
+ */
+int run_system_command(char *buffer)
+{
+    int pid, status;
+
+    if (!buffer || !buffer[0]) return 1;
+
+    /* fork child */
+    pid=fork();
+    if (pid == -1) {
+        /* failed */
+        g_printf("run_system_command(): fork() failed\n");
+        return 1;
+    }
+    else if (pid == 0) {
+	/* child */
+	int retval;
+	priv_drop(); /* we drop _all_ priviledges */
+        retval=system(buffer);	/* execute command */
+        _exit(retval);
+    }
+    else {
+	/* parent */
+        while (waitpid(pid, &status, WNOHANG) != pid) {
+            handle_signals();
+            if (iq.queued)
+                  do_queued_ioctl();
+            usleep(10000);
+        }
+        return WEXITSTATUS(status);
+    }
+}
