@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -25,7 +25,7 @@
  * Debug level.
  * 0 - normal / 1 - useful / 2 - too much
  */
-#define DEBUG_FATFS	0
+#define DEBUG_FATFS	2
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -59,6 +59,7 @@
 #include "emu.h"
 #include "disks.h"
 #include "fatfs.h"
+#include "doshelpers.h"
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -83,6 +84,7 @@ static unsigned next_cluster(fatfs_t *, unsigned);
 static void build_boot_blk(fatfs_t *);
 static void make_i1342_blk(unsigned char *, unsigned, unsigned, unsigned, unsigned);
 
+static char *bootfile = 0;
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 void fatfs_init(struct disk *dp)
@@ -494,6 +496,7 @@ void scan_dir(fatfs_t *f, unsigned oi)
         if(!strcasecmp(dent->d_name, "ibmbio.com")) f->sys_type |= 4;
         if(!strcasecmp(dent->d_name, "ibmdos.com")) f->sys_type |= 8;
         if(!strcasecmp(dent->d_name, "ipl.sys")) f->sys_type |= 0x10;
+        if(!strcasecmp(dent->d_name, "kernel.sys")) f->sys_type |= 0x20;
       }
       closedir(dir);
     }
@@ -510,12 +513,20 @@ void scan_dir(fatfs_t *f, unsigned oi)
       sf[1] = "ibmdos.com";
       scans = 3;
     }
-    if((f->sys_type & 0x10) == 0x10) {
-      f->sys_type = 0x10;	/* FreeDOS */
+    if((f->sys_type & 0x30) == 0x10) {
+      f->sys_type = 0x10;	/* FreeDOS, orig. Patv kernel */
       sf[0] = "ipl.sys";
       scans = 2;
     }
 
+    if((f->sys_type & 0x30) == 0x20) {
+      f->sys_type = 0x20;	/* FreeDOS, FD maintained kernel */
+      sf[0] = "kernel.sys";
+      scans = 2;
+      bootfile = malloc(strlen(name) + strlen(sf[0]) + 1);
+      sprintf(bootfile, "%s%s", name, sf[0]);
+    }
+
     fatfs_msg("system type is 0x%x\n", f->sys_type);
   }
 
@@ -614,7 +625,52 @@ void add_object(fatfs_t *f, unsigned parent, char *name)
   struct stat sb;
   obj_t tmp_o = {{0}, 0};
   unsigned u;
+  static int first = 1;
+  static int exxx = 0;
+  static char esys[16] = "";
+  static char ebat[16] = "";
 
+  if (first) {
+    first = 0;
+    if (config.emusys) {
+      snprintf(esys, 8+1+3+1,"config.%s", config.emusys);
+      esys[12] =0;
+      /* If we have to fake config.sys, it will _only_ be here
+       * and we don't need it after the switch to the lredir'ed drive.
+       * Hence we disable config.emusys.
+       */
+      config.emusys = 0;
+      exxx += 2;
+    }
+    if (config.emubat) {
+      snprintf(ebat, 8+1+3+1,"autoexec.%s", config.emubat);
+      ebat[12] =0;
+      exxx += 2;
+    }
+  }
+
+  if (!parent && exxx) {
+    if (esys[0] && !strcmp(name, "config.sys")) {
+      fatfs_deb("add_object: ignored real %s in favour of %s\n", name, esys);
+      exxx -= 1;
+      return;
+    }
+    if (ebat[0] && !strcmp(name, "autoexec.bat")) {
+      fatfs_deb("add_object: ignored real %s in favour of %s\n", name, ebat);
+      exxx -= 1;
+      return;
+    }
+    if (esys[0] && !strcmp(name, esys)) {
+      tmp_o.is.faked_sys = 1;
+      fatfs_deb("add_object: faked %s as DOS entry CONFIG.SYS\n", esys);
+      exxx -= 1;
+    }
+    else if (ebat[0] && !strcmp(name, ebat)) {
+      tmp_o.is.faked_bat = 1;
+      fatfs_deb("add_object: faked %s as DOS entry AUTOEXEC.BAT\n", ebat);
+      exxx -= 1;
+    }
+  }
   if(!(strcmp(name, ".") && strcmp(name, ".."))) return;
 
   if(!(s = full_name(f, parent, name))) {
@@ -687,6 +743,8 @@ unsigned make_dos_entry(fatfs_t *f, obj_t *o, unsigned char **e)
   memset(dos_ent, 0, sizeof dos_ent);
 
   s = o->name;
+  if (o->is.faked_sys) s = "config.sys";
+  if (o->is.faked_bat) s = "autoexec.bat";
 
   if(o->is.this_dir) {
     s = ".";
@@ -963,6 +1021,37 @@ unsigned next_cluster(fatfs_t *f, unsigned clu)
   return clu + 1;
 }
 
+/*
+ * This will be called by dos_helper (base/async/int.c)
+ * when the bootsector is executed.
+ * We load the systemfile directly and then jump to the entry point.
+ * A normal bootsector at 0x7c00 won't work, because the file will
+ * overwrite 0x7c00. Instead of fiddling with moving the bootsector
+ * 'out of danger' we just do the stuff here in 32-bit space.
+ */
+void fdkernel_boot_mimic(void)
+{
+  int f, size;
+  unsigned int loadaddress = 0x600;
+
+  if (!bootfile) {
+    error("BOOT-helper requested, but no systemfile available\n");
+    leavedos(99);
+  }
+  if ((f = open(bootfile, O_RDONLY)) == -1) {
+    error("cannot open DOS system file %s\n", bootfile);
+    leavedos(99);
+  }
+  size = lseek(f, 0, SEEK_END);
+  lseek(f, 0, SEEK_SET);
+  read(f, (void *)loadaddress, size);
+  close(f);
+  LWORD(cs) = LWORD(ds) = LWORD(es) = loadaddress >> 4;
+  LWORD(eip) = 0;
+  LWORD(ebx) = 0x80;	/* boot drive */
+  LWORD(ss) = 0x1FE0;
+  LWORD(esp) = 0x7c00;	/* temp stack */
+}
 
 /*
  * Build our own boot block (if no "boot.blk" file was found).
@@ -1092,13 +1181,25 @@ void build_boot_blk(fatfs_t *f)
       fatfs_msg("made boot block suitable for PC-, DR-, Open-, NOVELL-DOS\n");
       break;
 
-    case 0x10:			/* FreeDOS */
+    case 0x10:			/* FreeDOS, orig. Patv kernel */
       make_i1342_blk(d1, d_o, (f->obj[1].size + 0x1ff) >> 9, 0x2000, 0x0);
       d0[0x12] = 1;		/* 1 entry */
 
       d0[0x03] = 0x20;		/* start seg */
       d0[0x0a] = 0x80;		/* dl */
 
+      fatfs_msg("made boot block suitable for DosC\n");
+      break;
+
+    case 0x20:			/* FreeDOS, FD maintained kernel */
+			/* boot loading is done by DOSEMU-HELPER
+			   and above fdkernel_boot_mimic() function */
+      b[0x40] = 0xb8;	/* mov ax,0fdh */
+      b[0x41] = DOS_HELPER_BOOTSECT;
+      b[0x42] = 0x00;
+      b[0x43] = 0xcd;	/* int 0e6h */
+      b[0x44] = 0xe6;
+
       fatfs_msg("made boot block suitable for FreeDOS\n");
       break;
 
