@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2000 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -27,13 +27,7 @@
 
 #ifdef __linux__
 #include <signal.h>
-#if GLIBC_VERSION_CODE >= 2000
 #include <sys/vt.h>
-#else
-#include <linux/vt.h>
-#endif
-#include <linux/fd.h>
-#include <linux/hdreg.h>
 #include <sys/vm86.h>
 #include <syscall.h>
 #endif
@@ -170,7 +164,8 @@ io_select(fd_set fds)
 
     default:			/* has at least 1 descriptor ready */
 
-      if ((mice->intdrv || mice->type == MOUSE_PS2) && mice->fd >= 0)
+      if ((mice->intdrv || mice->type == MOUSE_PS2 || mice->type == MOUSE_IMPS2)
+	  && mice->fd >= 0)
 	if (FD_ISSET(mice->fd, &fds)) {
 		m_printf("MOUSE: We have data\n");
 	  pic_request(PIC_IMOUSE);
