@@ -1,27 +1,24 @@
+/* 
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* 	MS-DOS API translator for DOSEMU\'s DPMI Server
  *
  * DANG_BEGIN_MODULE msdos.h
  *
+ * REMARK
  * MS-DOS API translator allows DPMI programs to call DOS service directly
  * in protected mode.
  *
+ * /REMARK
  * DANG_END_MODULE
  *
  * First Attempted by Dong Liu,  dliu@rice.njit.edu
  *
- * $Log: msdos.h,v $
- * Revision 1.2  1995/05/06  16:25:48  root
- * Prep for 0.60.2.
- *
- * Revision 1.1  1995/04/08  22:31:40  root
- * Initial revision
- *
  */
 
-#ifndef lint
-static char *rcsid = "$Header: /usr/src/dosemu0.60/dpmi/RCS/msdos.h,v 1.2 1995/05/06 16:25:48 root Exp root $";
-#endif
-
 enum { ES_INDEX = 0, CS_INDEX = 1, SS_INDEX = 2,  DS_INDEX = 3,
        FS_INDEX = 4, GS_INDEX = 5 };
 
@@ -29,13 +26,8 @@ enum { ES_INDEX = 0, CS_INDEX = 1, SS_INDEX = 2,  DS_INDEX = 3,
 static struct vm86_regs SAVED_REGS;
 static struct vm86_regs MOUSE_SAVED_REGS;
 static struct vm86_regs VIDEO_SAVED_REGS;
-#define S_REG(reg) (SAVED_REGS.##reg)
-#endif
-#ifdef __NetBSD__
-static struct sigcontext SAVED_REGS;
-static struct sigcontext MOUSE_SAVED_REGS;
-static struct sigcontext VIDEO_SAVED_REGS;
-#define S_REG(reg) (SAVED_REGS.##reg)
+static struct vm86_regs INT15_SAVED_REGS;
+#define S_REG(reg) (SAVED_REGS.reg)
 #endif
 
 
@@ -78,6 +70,8 @@ static unsigned short CURRENT_ENV_SEL;
 static unsigned short PARENT_PSP;
 static unsigned short PARENT_ENV_SEL;
 static int in_dos_21 = 0;
+static int last_dos_21 = 0;
+
 static int old_dos_terminate(struct sigcontext_struct *scp)
 {
     REG(cs)  = CURRENT_PSP;
@@ -117,14 +111,14 @@ static int old_dos_terminate(struct sigcontext_struct *scp)
 		(Segments[i].limit > parent_off))
 		break;
 	if (i>=MAX_SELECTORS) {
-	    error("DPMI: WARNNING !!! unvalid termninate address found in PSP\n");
+	    error("DPMI: WARNING !!! invalid terminate address found in PSP\n");
 	    return 1;
 	} else
 	    _cs = (i << 3)|7;
 	_eip = parent_off;
 #endif	
 	/* set parent\'s PSP to itself, so dos won\'t free memory */
-	/* I believe winkernel put the selector of parent psp there */
+	/* I believe winkernel puts the selector of parent psp there */
 	PARENT_PSP = *(unsigned short *)((char *)(CURRENT_PSP<<4) + 0x16);
 	*(unsigned short *)((char *)(CURRENT_PSP<<4) + 0x16) = CURRENT_PSP;
 	
@@ -144,8 +138,12 @@ static int inline is_dos_selector(unsigned short sel)
 {
     unsigned long base = Segments[sel >> 3].base_addr;
 
-    if ( (base > 0xffff0) || (base & 0xf)) {
-      D_printf("DPMI: base adddress > 1MB or not para. aligned.\n");
+    if (base > 0x10ffef) {	/* ffff:ffff for DOS high */
+      D_printf("DPMI: base address %#lx of sel %#x > DOS limit\n", base, sel);
+      return 0;
+    } else
+    if (base & 0xf) {
+      D_printf("DPMI: base address %#lx of sel %#x not para. aligned.\n", base, sel);
       return 0;
     } else
       return 1;
@@ -168,9 +166,9 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 {
     /* only consider DOS and some BIOS services */
     switch (intr) {
-    case 0x10:
-    case 0x20 ... 0x21:
-    case 0x25 ... 0x26:
+    case 0x10: case 0x15:
+    case 0x20: case 0x21:
+    case 0x25: case 0x26:
     case 0x33:			/* mouse function */
 	break;
     case 0x2f:
@@ -190,8 +188,8 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 
     if (DTA_over_1MB && intr == 0x21 ) {
 	switch (_HI(ax)) {	/* functions use DTA */
-	case 0x11 ... 0x12:	/* find first/next using FCB */
-	case 0x4e ... 0x4f:	/* find first/next */
+	case 0x11: case 0x12:	/* find first/next using FCB */
+	case 0x4e: case 0x4f:	/* find first/next */
 	    memmove(DTA_under_1MB, DTA_over_1MB, 0x80);
 	    break;
 	}
@@ -219,7 +217,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 	    switch (_LO(ax)) {
 	    case 0x0:		/* user character load */
 	    case 0x10:		/* user specified character definition table */
-	    case 0x20 ... 0x21:
+	    case 0x20: case 0x21:
 		ES_MAPPED = 1;
 		break;
 	    default:
@@ -241,14 +239,18 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 	    return 0;
 	}
 	break;
+    case 0x15:			/* misc */
+	INT15_SAVED_REGS = REGS;
+	return 0;
     case 0x20:			/* DOS terminate */
 	return old_dos_terminate(scp);
     case 0x21:
 	if (in_dos_21) 
-	    dbug_printf("DPMI: int 0x21 called recursively in_dos_21=%d.\n",in_dos_21);
-	if (_HI(ax) != 0x4b && _HI(ax) != 0x48 && _HI(ax) != 0x49 &&
-	    _HI(ax) != 0x4a && _HI(ax) != 0x25 && _HI(ax) != 0x35 )
-	    in_dos_21++;
+	dbug_printf("DPMI: int21 AX=%#04x called recursively "
+		    "from inside %#04x, in_dos_21=%d\n",
+		_LWORD(eax), last_dos_21, in_dos_21);
+	last_dos_21 = _LWORD(eax);
+
 	SAVED_REGS = REGS;
 	READ_DS_COPIED = 0;
 	switch (_HI(ax)) {
@@ -325,20 +327,20 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 	case 0x19:		/* and segment register translation. */
 	case 0x2a ... 0x2e:
 	case 0x30 ... 0x34:
-	case 0x36 ... 0x37:
+	case 0x36: case 0x37:
 	case 0x3e:
 	case 0x42:
-	case 0x45 ... 0x46:
+	case 0x45: case 0x46:
 	case 0x4d:
 	case 0x4f:		/* find next */
 	case 0x54:
-	case 0x58 ... 0x59:
+	case 0x58: case 0x59:
 	case 0x5c:		/* lock */
 	case 0x66 ... 0x68:	
 	case 0xF8:		/* OEM SET vector */
+	    in_dos_21++;
 	    return 0;
 	case 0x00:		/* DOS terminate */
-	    in_dos_21--;
 	    return old_dos_terminate(scp);
 	case 0x09:		/* Print String */
 	    if ( !is_dos_selector(_ds)) {
@@ -355,6 +357,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 			break;
 		}
 	    }
+	    in_dos_21++;
 	    return 0;
 	case 0x0a:		/* buffered keyboard input */
 	case 0x38:
@@ -376,28 +379,32 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		/*memmove(DTA_under_1MB, DTA_over_1MB, 0x80);*/
 	    } else
 		DTA_over_1MB = 0;
+		in_dos_21++;
 	    return 0;
 	case 0x2f:		/* GET DTA */
 	    if (DTA_over_1MB) {
 		_es = USER_DTA_SEL;
 		_ebx = USER_DTA_OFF;
 		return 1;
-	    } else
+	    } else {
+		in_dos_21++;
 		return 0;
+	    }
 	/* FCB functions */	    
-	case 0x0f ... 0x10:	/* These are not supported by */
-	case 0x14 ... 0x15:	/* dosx.exe, according to Ralf Brown */
+	case 0x0f: case 0x10:	/* These are not supported by */
+	case 0x14: case 0x15:	/* dosx.exe, according to Ralf Brown */
 	case 0x21 ... 0x24:
-	case 0x27 ... 0x28:
+	case 0x27: case 0x28:
 	    _HI(ax) = 0xff;
 	    return 1;
-	case 0x11 ... 0x12:	/* find first/next using FCB */
+	case 0x11: case 0x12:	/* find first/next using FCB */
 	case 0x13:		/* Delete using FCB */
 	case 0x16:		/* Create usring FCB */
 	case 0x17:		/* rename using FCB */
 	    user_FCB = (void *)SEL_ADR(_ds, _edx);
 	    REG(ds) =	DPMI_private_data_segment+DPMI_private_paragraphs;
 	    memmove(SEG_ADR((void *), ds, dx), user_FCB, 0x50);
+	    in_dos_21++;
 	    return 0;
 	case 0x29:		/* Parse a file name for FCB */
 	    {
@@ -422,14 +429,16 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 			    0x50);
 		}
 	    }
+	    in_dos_21++;
 	    return 0;
 	case 0x44:		/* IOCTL */
 	    switch (_LO(ax)) {
 	    case 0x02 ... 0x05:
-	    case 0x0c ... 0x0d:
+	    case 0x0c: case 0x0d:
 		DS_MAPPED = 1;
 		break;
 	    default:
+		in_dos_21++;
 		return 0;
 	    }
 	    break;
@@ -439,6 +448,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		READ_DS_COPIED = 1;
 		S_REG(ds) = _ds;
 	    }
+	    in_dos_21++;
 	    return 0;
 	case 0x4b:		/* EXEC */
 	    D_printf("BCC: call dos exec.\n");
@@ -530,6 +540,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		    CURRENT_ENV_SEL = 0;
 	    }
 	    CURRENT_PSP = LWORD(ebx);
+	    in_dos_21++;
 	    return 0;
 	case 0x26:
 	case 0x55:		/* create PSP */
@@ -538,6 +549,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		Segments[LWORD(edx) >> 3].used &&
 		(Segments[LWORD(edx) >> 3].base_addr < 0xffff0))
 		REG(edx) = (long) GetSegmentBaseAddress(LWORD(edx)) >> 4;
+	    in_dos_21++;
 	    return 0;
 	case 0x39:		/* mkdir */
 	case 0x3a:		/* rmdir */
@@ -551,12 +563,15 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 	    if ((_HI(ax) == 0x4e) && (_ecx & 0x8))
 		D_printf("DPMI: MS-DOS try to find volume label\n");
 	    if ( !is_dos_selector(_ds)) {
-		D_printf("DPMI: passing ASCIIZ > 1MB to dos\n"); 
+		char *src, *dst;
 		REG(ds) = DPMI_private_data_segment+DPMI_private_paragraphs;
-		strcpy((void *)((REG(ds) << 4)+LWORD(edx)),
-		       (void *)GetSegmentBaseAddress(_ds)+
-		       (DPMIclient_is_32 ? _edx : (_LWORD(edx))));
+		src = (char *)(GetSegmentBaseAddress(_ds)+(DPMIclient_is_32 ? _edx : (_LWORD(edx))));
+		dst = (char *)((REG(ds) << 4)+LWORD(edx));
+		D_printf("DPMI: passing ASCIIZ > 1MB to dos %#x\n", (int)dst); 
+		D_printf("%#x: '%s'\n", (int)src, src);
+		strcpy(dst, src);
 	    }
+	    in_dos_21++;
 	    return 0;
 	case 0x3f:		/* dos read */
 	    if ( !is_dos_selector(_ds)) {
@@ -566,6 +581,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		READ_DS_COPIED = 1;
 	    } else
 		READ_DS_COPIED = 0;
+	    in_dos_21++;
 	    return 0;
 	case 0x40:		/* DOS Write */
 	    if ( !is_dos_selector(_ds)) {
@@ -579,6 +595,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		READ_DS_COPIED = 1;
 	    } else
 		READ_DS_COPIED = 0;
+	    in_dos_21++;
 	    return 0;
 	case 0x53:		/* Generate Drive Parameter Table  */
 	    {
@@ -603,6 +620,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 			    0x60);
 		}
 	    }
+	    in_dos_21++;
 	    return 0;
 	case 0x56:		/* rename file */
 	    {
@@ -623,10 +641,13 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 			    (DPMIclient_is_32 ? _edi : (_LWORD(edi))));
 		}
 	    }
+	    in_dos_21++;
 	    return 0;
 	case 0x57:		/* Get/Set File Date and Time Using Handle */
-	    if ((_LO(ax) == 0) || (_LO(ax) == 1))
+	    if ((_LO(ax) == 0) || (_LO(ax) == 1)) {
+		in_dos_21++;
 		return 0;
+	    }
 	    ES_MAPPED = 1;
 	    break;
 	case 0x5e:
@@ -637,7 +658,8 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 	    break;
 	case 0x5f:		/* redirection */
 	    switch (_LO(ax)) {
-	    case 0 ... 1:
+	    case 0: case 1:
+		in_dos_21++;
 		return 0;
 	    case 2 ... 6:
 		REG(ds) = DPMI_private_data_segment + DPMI_private_paragraphs;
@@ -655,6 +677,7 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 			(void *)GetSegmentBaseAddress(_es) +
 			(DPMIclient_is_32 ? _edi : (_LWORD(edi))),
 			0x100);
+		in_dos_21++;
 		return 0;
 	    }
 	case 0x60:		/* Get Fully Qualified File Name */
@@ -673,15 +696,19 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 		    (void *)GetSegmentBaseAddress(_es) +
 		    (DPMIclient_is_32 ? _edi : (_LWORD(edi))),
 		    0x100);
+	    in_dos_21++;
 	    return 0;
 	case 0x6c:		/*  Extended Open/Create */
 	    if ( !is_dos_selector(_ds)) {
-		D_printf("DPMI: passing ASCIIZ > 1MB to dos\n"); 
+		char *src, *dst;
 		REG(ds) = DPMI_private_data_segment+DPMI_private_paragraphs;
-		strcpy((void *)((REG(ds) << 4)+LWORD(esi)),
-		       (void *)GetSegmentBaseAddress(_ds)+
-		       (DPMIclient_is_32 ? _esi : (_LWORD(esi))));
+		src = (char *)(GetSegmentBaseAddress(_ds)+(DPMIclient_is_32 ? _esi : (_LWORD(esi))));
+		dst = (char *)((REG(ds) << 4)+LWORD(esi));
+		D_printf("DPMI: passing ASCIIZ > 1MB to dos %#x\n", (int)dst); 
+		D_printf("%#x: '%s'\n", (int)src, src);
+		strcpy(dst, src);
 	    }
+	    in_dos_21++;
 	    return 0;
 	default:
 	    break;
@@ -732,29 +759,40 @@ msdos_pre_extender(struct sigcontext_struct *scp, int intr)
 
     if (DS_MAPPED) {
 	if ( !is_dos_selector(_ds)) {
-	    D_printf("DPMI: whole segment of DS copy to DOS.\n");
+	    char *src, *dst;
+	    int len;
 	    DS_MAPPED = _ds;
 	    REG(ds) =   DPMI_private_data_segment+DPMI_private_paragraphs;
-	    memmove ((void *)(REG(ds)<<4), (void *)GetSegmentBaseAddress(_ds),
-		    ((Segments[_ds >> 3].limit > 0xffff) ||
-		     Segments[_ds >> 3].is_big) ?
-		    0xffff : Segments[_ds >> 3].limit);
+	    src = (char *)GetSegmentBaseAddress(_ds);
+	    dst = (char *)(REG(ds)<<4);
+	    len = ((Segments[_ds >> 3].limit > 0xffff) ||
+	    	Segments[_ds >> 3].is_big) ?
+		0xffff : Segments[_ds >> 3].limit;
+	    D_printf("DPMI: whole segment of DS at %#x copy to DOS at %#x for %#x\n",
+		(int)src, (int)dst, len);
+	    memmove (dst, src, len);
 	} else
 	    DS_MAPPED = 0;
     }
 
     if (ES_MAPPED) {
 	if ( !is_dos_selector(_es)) {
-	    D_printf("DPMI: whole segment of ES copy to DOS.\n");
+	    char *src, *dst;
+	    int len;
 	    ES_MAPPED = _es;
 	    REG(es) =   DPMI_private_data_segment+DPMI_private_paragraphs;
-	    memmove ((void *)(REG(es)<<4), (void *)GetSegmentBaseAddress(_es),
-		    ((Segments[_es >> 3].limit > 0xffff)||
-		     Segments[_ds >> 3].is_big) ?
-		    0xffff : Segments[_es >> 3].limit);
+	    src = (char *)GetSegmentBaseAddress(_es);
+	    dst = (char *)(REG(es)<<4);
+	    len = ((Segments[_es >> 3].limit > 0xffff) ||
+	    	Segments[_es >> 3].is_big) ?
+		0xffff : Segments[_es >> 3].limit;
+	    D_printf("DPMI: whole segment of ES at %#x copy to DOS at %#x for %#x\n",
+		(int)src, (int)dst, len);
+	    memmove (dst, src, len);
 	} else
 	    ES_MAPPED = 0;
     }
+    if (intr==0x21) in_dos_21++;
     return 0;
 }
 
@@ -785,8 +823,8 @@ msdos_post_exec()
 /*
  * DANG_BEGIN_FUNCTION msdos_post_extender
  *
- * This function is called after retrun from real mode DOS service
- * All real mode segment register is changed to protected mode selector
+ * This function is called after return from real mode DOS service
+ * All real mode segment registers are changed to protected mode selectors
  * And if client\'s data buffer is above 1MB, necessary buffer copying
  * is performed.
  *
@@ -806,19 +844,19 @@ msdos_post_extender(int intr)
 	} else
 	    break;
     case 0x15:
-      switch(HI(ax)) {
-        case 0xc0:      /* Get Configuartion */
+	/* we need to save regs at int15 because AH has the return value */
+	if ((INT15_SAVED_REGS.eax & 0xff00) == 0xc000) { /* Get Configuration */
                 if (REG(eflags)&CF)
                         return;
                 if (!(dpmi_stack_frame[current_client].es =
-                         ConvertSegmentToDescriptor(REG(es))))
-                return;
-        default:
-                return;
+                         ConvertSegmentToDescriptor(REG(es)))) return;
+                break;
       }
+        else
+                return;
     /* only copy buffer for dos services */
     case 0x21:
-    case 0x25 ... 0x26:
+    case 0x25: case 0x26:
     case 0x33:			/* mouse */
 	break;
     default:
@@ -827,8 +865,8 @@ msdos_post_extender(int intr)
 
     if (DTA_over_1MB && intr == 0x21 ) {
 	switch (S_HI(ax)) {	/* functions use DTA */
-	case 0x11 ... 0x12:	/* find first/next using FCB */
-	case 0x4e ... 0x4f:	/* find first/next */
+	case 0x11: case 0x12:	/* find first/next using FCB */
+	case 0x4e: case 0x4f:	/* find first/next */
 	    memmove(DTA_over_1MB, DTA_under_1MB, 0x80);
 	    break;
 	}
@@ -836,23 +874,33 @@ msdos_post_extender(int intr)
 
     if (DS_MAPPED) {
 	unsigned short my_ds;
+	char *src, *dst;
+	int len;
 	my_ds =   DPMI_private_data_segment+DPMI_private_paragraphs;
-	memmove ((void *)GetSegmentBaseAddress(DS_MAPPED), (void *)(my_ds<<4), 
-		((Segments[DS_MAPPED >> 3].limit > 0xffff)||
-		 Segments[DS_MAPPED >> 3].is_big) ?
-		0xffff : Segments[DS_MAPPED >> 3].limit);
+	src = (char *)(my_ds<<4);
+	dst = (char *)GetSegmentBaseAddress(DS_MAPPED);
+	len = ((Segments[DS_MAPPED >> 3].limit > 0xffff) ||
+	    	Segments[DS_MAPPED >> 3].is_big) ?
+		0xffff : Segments[DS_MAPPED >> 3].limit;
+	D_printf("DPMI: DS_MAPPED seg at %#x copy back at %#x for %#x\n",
+		(int)src, (int)dst, len);
+	memmove (dst, src, len);
     } 
-    DS_MAPPED = 0;
 
     if (ES_MAPPED) {
 	unsigned short my_es;
+	char *src, *dst;
+	int len;
 	my_es =   DPMI_private_data_segment + DPMI_private_paragraphs;
-	memmove ((void *)GetSegmentBaseAddress(ES_MAPPED), (void *)(my_es<<4), 
-		((Segments[ES_MAPPED >> 3].limit > 0xffff) ||
-		Segments[ES_MAPPED >> 3].is_big) ?
-		 0xffff : Segments[ES_MAPPED >> 3].limit);
+	src = (char *)(my_es<<4);
+	dst = (char *)GetSegmentBaseAddress(ES_MAPPED);
+	len = ((Segments[ES_MAPPED >> 3].limit > 0xffff) ||
+	    	Segments[ES_MAPPED >> 3].is_big) ?
+		0xffff : Segments[ES_MAPPED >> 3].limit;
+	D_printf("DPMI: ES_MAPPED seg at %#x copy back at %#x for %#x\n",
+		(int)src, (int)dst, len);
+	memmove (dst, src, len);
     } 
-    ES_MAPPED = 0;
 	       
     switch (intr) {
     case 0x21:
@@ -867,14 +915,14 @@ msdos_post_extender(int intr)
 		 GetSegmentBaseAddress(dpmi_stack_frame[current_client].cs)>>4;
 	    /* restore parent PSP */
 	    *(unsigned short *)((char *)(CURRENT_PSP<<4) + 0x16) = PARENT_PSP;
-	    return;
+	    break;
 	case 0x09:		/* print String */
 	case 0x1a:		/* set DTA */
 	    dpmi_stack_frame[current_client].edx = S_REG(edx);
-	    return;
-	case 0x11 ... 0x12:	/* findfirst/next using FCB */
+	    break;
+	case 0x11: case 0x12:	/* findfirst/next using FCB */
 	    memmove(user_FCB, SEG_ADR((void *), ds, dx), 0x50);
-	    return ;
+	    break ;
 	case 0x29:		/* Parse a file name for FCB */
 	    {
 		unsigned short seg =
@@ -895,17 +943,24 @@ msdos_post_extender(int intr)
 			    (void *)((REG(es)<<4) + LWORD(edi)),  0x50);
 		}
 	    }
-	    return;
+	    break;
 	case 0x2f:		/* GET DTA */
 	case 0x34:		/* Get Address of InDOS Flag */
 	case 0x35:		/* GET Vector */
 	case 0x52:		/* Get List of List */
-        case 0x59:		/* Get EXTENDED ERROR INFORMATION */
-	    dpmi_stack_frame[current_client].es =
+	    if (!ES_MAPPED)
+	      dpmi_stack_frame[current_client].es =
 	               ConvertSegmentToDescriptor(REG(es));
 	    break;
+
+        case 0x59:		/* Get EXTENDED ERROR INFORMATION */
+	    if(LWORD(eax) == 0x22 && !ES_MAPPED) { /* only this code has a pointer */
+		dpmi_stack_frame[current_client].es =
+			ConvertSegmentToDescriptor(REG(es));
+	    }
+	    break;
 	case 0x38:
-	    if (S_LO(ax) == 0x00) { /* get contry info */
+	    if (S_LO(ax) == 0x00 && !DS_MAPPED) { /* get country info */
 		dpmi_stack_frame[current_client].ds =
 	               ConvertSegmentToDescriptor(REG(ds));
 	    }
@@ -953,7 +1008,7 @@ msdos_post_extender(int intr)
 #endif 		
 		dpmi_stack_frame[current_client].ebx = ConvertSegmentToDescriptor(psp);
 	    }
-	    return;
+	    break;
 	case 0x53:		/* Generate Drive Parameter Table  */
 	    {
 		unsigned short seg =
@@ -969,21 +1024,21 @@ msdos_post_extender(int intr)
 			    0x60);
 		}
 	    }
-	    return ;
+	    break ;
 	case 0x56:		/* rename */
 	    dpmi_stack_frame[current_client].edx = S_REG(edx);
 	    dpmi_stack_frame[current_client].edi = S_REG(edi);
-	    return ;
+	    break ;
 	case 0x5d:
 	    if (S_LO(ax) == 0x06) /* get address of DOS swappable area */
 				/*        -> DS:SI                     */
 		dpmi_stack_frame[current_client].ds = ConvertSegmentToDescriptor(REG(ds));
-	    return;
+	    break;
 	case 0x3f:
 	    if (READ_DS_COPIED) {
 		dpmi_stack_frame[current_client].edx = S_REG(edx);
 		if (LWORD(eflags) & CF)
-		    return;
+		    break;
 		memmove((void *)(GetSegmentBaseAddress
 				(dpmi_stack_frame[current_client].ds)
 			 + (DPMIclient_is_32 ? S_REG(edx) : (S_LWORD(edx)))),
@@ -998,8 +1053,8 @@ msdos_post_extender(int intr)
 	    break;
 	case 0x5f:		/* redirection */
 	    switch (S_LO(ax)) {
-	    case 0 ... 1:
-		return ;
+	    case 0: case 1:
+		break ;
 	    case 2 ... 6:
 		dpmi_stack_frame[current_client].esi = S_REG(esi);
 		memmove ((void *)GetSegmentBaseAddress
@@ -1016,7 +1071,7 @@ msdos_post_extender(int intr)
 	    }
 	    break;
 	default:
-	    return;
+	    break;
 	}
 	break;
     case 0x25:			/* Absolute Disk Read */
@@ -1034,23 +1089,25 @@ msdos_post_extender(int intr)
 	      (dpmi_stack_frame[current_client].esp&0xffff)) =
 		 LWORD(eflags);
 	}
-	return;
+	break;
     case 0x33:			/* mouse */
 	switch (MOUSE_SAVED_REGS.eax & 0xffff) {
 	case 0x14:		/* swap call back */
 	    dpmi_stack_frame[current_client].es =
                   	    ConvertSegmentToDescriptor(REG(es)); 
-	    return;
+	    break;
 	case 0x19:		/* Get User Alternate Interrupt Address */
 	    dpmi_stack_frame[current_client].ebx =
                   	    ConvertSegmentToDescriptor(LWORD(ebx)); 
-	    return;
+	    break;
 	default:
-	    return;
+	    break;
 	}
     default:
-	return;
+	break;
     }
+    DS_MAPPED = 0;
+    ES_MAPPED = 0;
 }
 
 static char decode_use_16bit;
@@ -1148,13 +1205,16 @@ check_prefix (struct sigcontext_struct *scp)
 	use_prefix = 1;
 	break;
     default:
-    	D_printf("DPMI: check_prefix not covered for *csp=%x, return=NULL\n", *csp);
+	break;
+    }
+    if (use_prefix) {
+    	D_printf("DPMI: check_prefix covered for *csp=%x\n", *csp);
     }
     return prefix;
 }
 /*
- * this function try to decode opcode 0x8e (mov Sreg,m/r16), return
- * the length of the instructon.
+ * this function tries to decode opcode 0x8e (mov Sreg,m/r16), returns
+ * the length of the instruction.
  */
 
 static int
@@ -1165,8 +1225,6 @@ decode_8e(struct sigcontext_struct *scp, unsigned short *src,
     unsigned char mod, rm, reg;
     int len = 0;
 
-    if (!decode_use_16bit)	/*  32bit decode not implmented yet*/
-	return 0;
     csp = (unsigned char *) SEL_ADR(_cs, _eip);
 
     prefix = check_prefix(scp);
@@ -1250,8 +1308,6 @@ decode_load_descriptor(struct sigcontext_struct *scp, unsigned short
     unsigned offset;
     int len = 0;
 
-    if (!decode_use_16bit)	/*  32bit decode not implmented yet*/
-	return 0;
     csp = (unsigned char *) SEL_ADR(_cs, _eip);
 
     prefix = check_prefix(scp);
@@ -1338,7 +1394,7 @@ decode_load_descriptor(struct sigcontext_struct *scp, unsigned short
 
 /*
  * decode_modify_segreg_insn tries to decode instructions which would modify a
- * segment register, return the length of the insn.
+ * segment register, returns the length of the insn.
  */
 static  int
 decode_modify_segreg_insn(struct sigcontext_struct *scp, unsigned
@@ -1370,15 +1426,18 @@ decode_modify_segreg_insn(struct sigcontext_struct *scp, unsigned
 	    decode_use_16bit = 1;
     }
 	
-    /* first try mov sreg, .. */
+    /* first try mov sreg, .. (equal for 16/32 bit operand size) */
     if ((len = decode_8e(scp, segment, sreg))) {
-	_eip = old_eip;
-	return len + size_prfix;
+      _eip = old_eip;
+      return len + size_prfix;
     }
-    /* then try lds, les ... */
-    if ((len = decode_load_descriptor(scp, segment, sreg))) {
+ 
+    if (decode_use_16bit) {	/*  32bit decode not implemented yet */
+      /* then try lds, les ... */
+      if ((len = decode_load_descriptor(scp, segment, sreg))) {
 	_eip = old_eip;
 	return len+size_prfix;
+      }
     }
 
     /* then try pop sreg */
@@ -1415,7 +1474,7 @@ decode_modify_segreg_insn(struct sigcontext_struct *scp, unsigned
 	case 0xa9:		/* pop gs */
 	    len = 2;
 	    *segment = *ssp;
-	    *sreg = FS_INDEX;
+	    *sreg = GS_INDEX;
 	    _esp += decode_use_16bit ? 2 : 4;
 	break;
 	}
@@ -1451,77 +1510,27 @@ static  int msdos_fix_cs_prefix (struct sigcontext_struct *scp)
 }
 #endif
 
-/* see if client wants to access control registers */
-static int check_control_register_access(struct sigcontext_struct *scp)
-{
-    unsigned char *csp;
-    unsigned char mod, rm, reg;
-
-    csp = (unsigned char *) SEL_ADR(_cs, _eip);
-
-    if (*csp++ != 0x0f)
-	return 0;
-
-    if (*csp == 0x20) {	/* MOV r32,CR0/CR2/CR3 */
-	csp++;
-	mod = (*csp>>6) & 3;
-	reg = (*csp>>3) & 7;
-	rm = *csp & 0x7;
-	if (mod != 3)		/* mod must be 11 */
-	    return 0;
-	if (reg > 3)		/* only cr0 to cr3 */
-	    return 0;
-	_eip += 3;
-	if (reg != 0)		/* don\'t know content of cr1 .. cr3 */
-	    return 1;
-	switch (rm) {
-	case 0:
-	    _eax = 0x8000001f;	/* set all bits */
-	    break;
-	case 1:
-	    _ecx = 0x8000001f;
-	    break;
-	case 2:
-	    _edx = 0x8000001f;
-	    break;
-	case 3:
-	    _ebx = 0x8000001f;
-	    break;
-	case 4:
-	    _esp = 0x8000001f;
-	    break;
-	case 5:
-	    _ebp = 0x8000001f;
-	    break;
-	case 6:
-	    _esi = 0x8000001f;
-	    break;
-	case 7:
-	    _edi = 0x8000001f;
-	    break;
-	}
-	return 1;
-    } else if (*csp++ == 0x22) { /* MOV CR0/CR2/CR3,r32 */
-	_eip += 3;
-	return 1;		/* do nothing */
-    }
-    return 0;
-}
 
 static int msdos_fault(struct sigcontext_struct *scp)
 {
     unsigned char reg;
     unsigned short segment, desc;
     unsigned long len;
+    extern int cpu_trap_0f (unsigned char *, struct sigcontext_struct *);
+
+    D_printf("DPMI: msdos_fault, err=%#lx\n",_err);
 
     if ((_err & 0xffff) == 0) {	/*  not a selector error */
 	char fixed = 0;
 	unsigned char * csp;
 
-	if (check_control_register_access(scp))
-	    return 1;
-	
 	csp = (unsigned char *) SEL_ADR(_cs, _eip);
+
+	/* see if client wants to access control registers */
+	if (*csp == 0x0f) {
+	  if (cpu_trap_0f(csp, scp)) return 1;	/* 1=handled */
+	}
+	
 	switch (*csp) {
 	case 0x2e:		/* cs: */
 	    break;		/* do nothing */
@@ -1552,8 +1561,8 @@ static int msdos_fault(struct sigcontext_struct *scp)
 	case 0xf3:		/* REP, REPE */
 	    /* this might be a string insn */
 	    switch (*(csp+1)) {
-	    case 0xaa ... 0xab:		/* stos */
-	    case 0xae ... 0xaf:	        /* scas */
+	    case 0xaa: case 0xab:		/* stos */
+	    case 0xae: case 0xaf:	        /* scas */
 		/* only use es */
 		if (_es == 0) {
 		    D_printf("DPMI: client tries to use use gdt 0 as es\n");
@@ -1561,8 +1570,8 @@ static int msdos_fault(struct sigcontext_struct *scp)
 		    fixed = 1;
 		}
 		break;
-	    case 0xa4 ... 0xa5:		/* movs */
-	    case 0xa6 ... 0xa7:         /* cmps */
+	    case 0xa4: case 0xa5:		/* movs */
+	    case 0xa6: case 0xa7:         /* cmps */
 		/* use both ds and es */
 		if (_es == 0) {
 		    D_printf("DPMI: client tries to use use gdt 0 as es\n");
@@ -1578,7 +1587,7 @@ static int msdos_fault(struct sigcontext_struct *scp)
 	    }
 	    break;
 	case 0x3e:		/* ds: */
-	default:		/* assmue default is using ds, but if the */
+	default:		/* assume default is using ds, but if the */
 				/* client sets ss to 0, it is totally broken */
 	    if (_ds == 0) {
 		D_printf("DPMI: client tries to use use gdt 0 as ds\n");
@@ -1600,7 +1609,7 @@ static int msdos_fault(struct sigcontext_struct *scp)
     D_printf("DPMI: try mov to a invalid selector 0x%04x\n", segment);
 
 #if 1
-    /* olny allow using some special GTD\'s */
+    /* only allow using some special GTD\'s */
     if ((segment != 0x0040) && (segment != 0xa000) &&
 	(segment != 0xb000) && (segment != 0xb800) &&
 	(segment != 0xc000) && (segment != 0xe000) && (segment != 0xf000))
