@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -301,15 +301,13 @@ umb_allocate(int size)
 {
   int i;
 
-  if (size == 0)
-    return ((vm_address_t) 0);
-
   for (i = 0; i < UMBS; i++) {
     if (umbs[i].in_use && umbs[i].free) {
       if (umbs[i].size > size) {
 	int new_umb = umb_find_unused();
 
 	if (new_umb != UMB_NULL) {
+	 if(size) {
 	  umbs[new_umb].in_use = TRUE;
 	  umbs[new_umb].free = TRUE;
 	  umbs[new_umb].addr =
@@ -318,7 +316,8 @@ umb_allocate(int size)
 	    umbs[i].size - size;
 	  umbs[i].size = size;
 	  umbs[i].free = FALSE;
-	  return (umbs[i].addr);
+	 }
+	 return (umbs[i].addr);
 	}
       }
       else
@@ -544,16 +543,19 @@ xms_control(void)
 #define state (&_regs)
   case XMS_ALLOCATE_UMB:
     {
-      int size = WORD(state->edx) * 16;
-      vm_address_t addr = size ? umb_allocate(size) : (vm_address_t) 0;
+      int size = WORD(state->edx) << 4;
+      vm_address_t addr = umb_allocate(size);
 
       Debug0((dbg_fd, "Allocate UMB memory: 0x%04x\n",
 	      (unsigned) WORD(state->edx)));
       if (addr == (vm_address_t) 0) {
+        int avail=umb_query();
+
 	Debug0((dbg_fd, "Allocate UMB Failure\n"));
 	SETWORD(&(state->eax), 0);
-	SETLOW(&(state->ebx), 0xb0);
-	SETWORD(&(state->edx), umb_query() >> 4);
+	if(avail) SETLOW(&(state->ebx), 0xb0);
+	else SETLOW(&(state->ebx), 0xb1);
+	SETWORD(&(state->edx), avail >> 4);
       }
       else {
 	Debug0((dbg_fd, "Allocate UMB Success\n"));
@@ -620,7 +622,7 @@ FindFreeHandle(int start)
   /* first free handle is 1 */
   for (i = start; (i <= NUM_HANDLES) && (h == 0); i++) {
     if (!handles[i].valid) {
-      x_printf("\nXMS: found free handle: %d\n", i);
+      x_printf("XMS: found free handle: %d\n", i);
       h = i;
       break;
     }
@@ -934,8 +936,8 @@ xms_realloc_EMB(int api)
   else
     handles[h].size = REG(ebx) * 1024;
 
-  x_printf((api == OLDXMS) ? "XMS realloc EMB(old) %d to size 0x%04x\n" :
-	   "XMS realloc EMB(new) %d to size 0x%08x\n",
+  x_printf((api == OLDXMS) ? "XMS realloc EMB(old) %d to size 0x%04lx\n" :
+	   "XMS realloc EMB(new) %d to size 0x%08lx\n",
 	   h, handles[h].size);
 
   handles[h].addr = malloc(handles[h].size);
