@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -23,16 +23,16 @@
 #include <sys/time.h>
 #include <sys/ioctl.h>
 #include <net/route.h>
-#if defined (__GLIBC__) && __GLIBC__ >= 2
+#if GLIBC_VERSION_CODE >= 2000
   #include<netipx/ipx.h>
 #else
-  #include <linux/sockios.h>
-  #include <linux/ipx.h>
+  #include "Linux/ipx.h"
 #endif
 #include <netinet/in.h>
 #include <errno.h>
 
 #include "emu.h"
+#include "dosemu_select.h"
 
 #define FALSE   0
 #define TRUE    1
