@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -9,7 +9,7 @@
  *
  * Description: New port handling code for DOSEMU
  * 
- * Maintainers: Alberto Vignani (vignani@mbox.vol.it)
+ * Maintainers: Alberto Vignani (vignani@mail.tin.it)
  *
  * REMARK
  * This is the code that allows and disallows port access within DOSEMU.
@@ -33,17 +33,6 @@
 #include <errno.h>
 #include <malloc.h>
 #include <string.h>
-#ifdef __NetBSD__
-#include <stdio.h>
-#include <termios.h>
-#include <stdlib.h>
-#include <time.h>
-#include <ctype.h>
-#include <sys/times.h>
-#include <sys/time.h>
-#include <sys/types.h>
-#include <sys/times.h>
-#endif
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #if GLIBC_VERSION_CODE >= 2000
@@ -128,6 +117,7 @@ void SET_HANDLE_COND(int p, int h)
 /* ---------------------------------------------------------------------- */
 /* PORT TRACING								  */
 
+#if 0
 static long nyb2bin[16] =
 {
 	0x30303030, 0x31303030, 0x30313030, 0x31313030,
@@ -146,6 +136,7 @@ static char *
 
 	return s + 3;
 }
+#endif
 
 #define PORTLOG_MAXBITS		16
 #define PORTLOG_MASK		(~(-1 << PORTLOG_MAXBITS))
@@ -633,11 +624,22 @@ int port_rep_outd(ioport_t port, Bit32u *base, int df, Bit32u count)
  * I don't know what to do of this stuff... it was added incrementally to
  * port.c and has mainly to do with video code. This is not the right
  * place for it...
- * Anyway, this implements some HGC stuff for X
+ * Anyway, this implements some HGC stuff for X and the emuretrace
+ * port access for 0x3c0/0x3da
  *
  * SIDOC_END_FUNCTION
  */
 
+static int r3da_pending = 0;
+
+void do_r3da_pending (void)
+{
+  if (r3da_pending) {
+    (void)std_port_inb(r3da_pending);
+    r3da_pending = 0;
+  }
+}
+
 static Bit8u special_port_inb(ioport_t port)
 {
 	Bit8u res = 0xff;
@@ -655,7 +657,11 @@ static Bit8u special_port_inb(ioport_t port)
 	    }
 	}
 	if ((port==0x3ba)||(port==0x3da)) {
-		res = Attr_get_input_status_1();
+		res = Misc_get_input_status_1();
+		if (!config.usesX && !r3da_pending && (config.emuretrace>1)) {
+			set_ioperm(0x3c0,1,0);
+			r3da_pending = port;
+		}
 	}
 	else
 	if (port==0x3db)	/* light pen strobe reset */
@@ -687,6 +693,32 @@ static void special_port_outb(ioport_t port, Bit8u byte)
 		static int last_index = -1;
 		static int flip_flop = 1;
 
+		/* SIDOC_BEGIN_REMARK
+		 *
+		 * This is the core of the new emuretrace algorithm:
+		 * If a read of port 0x3da is performed we just set it
+		 *  as pending and set ioperm OFF for port 0x3c0
+		 * When a write to port 0x3c0 is then trapped, we perform
+		 *  any pending read to 0x3da and reset the ioperm for
+		 *  0x3c0 in the default ON state.
+		 * This way we avoid extra port accesses when the program
+		 * is only looking for the sync bits, and we don't miss
+		 * the case where the read to 0x3da is used to reset the
+		 * index/data flipflop for port 0x3c0. Futher accesses to
+		 * port 0x3c0 are handled at full speed.
+		 *
+		 * SIDOC_END_REMARK
+		 */
+		if (config.vga && !config.usesX && (config.emuretrace>1)) {
+		    if (r3da_pending) {
+			(void)std_port_inb(r3da_pending);
+			r3da_pending = 0;
+			set_ioperm(0x3c0,1,1);
+			port_real_outb(0x3c0, byte);
+			return;
+		    }
+		    goto defout;
+		}
 		flip_flop = !flip_flop;
 		if (flip_flop) {
 		/* JES This was last_index = 0x10..... WRONG? */
@@ -1203,8 +1235,7 @@ set_ioperm(int start, int size, int flag)
 	tmp = DOS_SYSCALL(ioperm(start, size, flag));
 	leave_priv_setting();
 
-	i_printf ("PORT: set_ioperm [%4x:%2d:%d] returns %d\n",start,size,flag,tmp);
-
+	i_printf ("nPORT: set_ioperm [%4x:%2d:%d] returns %d\n",start,size,flag,tmp);
 	return tmp;
 }
 
@@ -1212,7 +1243,3 @@ set_ioperm(int start, int size, int flag)
 /* ====================================================================== */
 
 
-#ifdef __NetBSD__
-#error please use old port code for NetBSD
-#endif				/* __NetBSD__ */
-
