@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -13,11 +13,6 @@
 #include <sys/kd.h>
 #include <sys/vt.h>
 #endif
-#ifdef __NetBSD__
-#include <machine/pcvt_ioctl.h>
-/* XXX duration is top half, pitch bottom. both zero results in no noise? */
-#define KIOCSOUND KDMKTONE		
-#endif
 
 #include "emu.h"
 #include "termio.h"
