@@ -1,9 +1,18 @@
+/* 
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * dualmon.c
  *
+ * REMARK
  * Dual-monitor support.
+ * /REMARK
  *
- * (C) 1994 under GPL, Hans Lermen (lermen@elserv.ffm.fgan.de)
+ * (C) 1994 under GPL, Hans Lermen (lermen@fgan.de)
+ * (put under DOSEMU-policy 1998, --Hans)
  *
  *          HGC-port-IO code is taken from hgc.c
  *          hgc.c is
@@ -50,9 +59,9 @@
  */
 
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <sys/mman.h>
 #include <errno.h>
 #include <malloc.h>
 
@@ -65,6 +74,7 @@
 #include "vc.h"
 #include "vga.h"
 #include "hgc.h"
+#include "mapping.h"
 
 extern void close_kmem(), open_kmem();
 
@@ -129,15 +139,14 @@ static int map_MDA_for_dualmon()
          || _IS_VS(Video_console)
        ) && (!_IS_VS(Video_hgc))) {
     int size=TEXT_SIZE;
-    open_kmem();
-    if (mmap((caddr_t) MDA_PHYS_TEXT_BASE, (size_t) size,
-         PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, mem_fd, MDA_PHYS_TEXT_BASE) == (caddr_t) -1) {
-      error("ERROR: mmap error in MDA video mapping %s\n", strerror(errno));
+    if (mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
+         (caddr_t) MDA_PHYS_TEXT_BASE, (size_t) size, PROT_READ | PROT_WRITE,
+         (void *)MDA_PHYS_TEXT_BASE) == (caddr_t) -1) {
+      error("mmap error in MDA video mapping %s\n", strerror(errno));
       return 0;
     }
     v_printf("VID: for dualmon, mapped MDA video ram at 0x%05x, size=0x%04x\n",
           MDA_PHYS_TEXT_BASE,size);                                 
-    close_kmem();
     return 1;
   }
   if (config.dualmon && (_IS_VS(Video_graphics) /* || _IS_VS(Video_console )*/ )) return 2;
@@ -266,6 +275,18 @@ static void MDA_init()
   port_out(0x28,0x03b8);
 } 
 
+static void reinit_MDA_regs()
+{
+  /* some video BIOSes are disabling MDA after they were initialised
+   * (such as on S3). We have to re-init some registers
+   */
+
+  /* Graphics allowed with 1 page (half mode) */
+  port_out(1,0x03bf);
+  /* Texmode, screen & cursor visible, page 0 */
+  port_out(0x28,0x03b8);
+} 
+
 static int dualmon_init(void)
 {
 v_printf("VID: dualmon_init called\n");
@@ -287,6 +308,7 @@ static void dualmon_close(void)
 static int dualmon_setmode(int type, int xsize,int ysize)
 {
   if (type==7) {
+    if (config.dualmon == 2) reinit_MDA_regs();
     Video->is_mapped = 1;
     WRITE_WORD(BIOS_CONFIGURATION, READ_WORD(BIOS_CONFIGURATION) | 0x30);
     WRITE_WORD(BIOS_CURSOR_SHAPE, 0x0b0d);
