@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -38,33 +38,15 @@
 #include <ctype.h>
 #include <termios.h>			/* needed for mouse.h */
 #include <sys/time.h>
-#ifdef __NetBSD__
-#include <sys/mman.h>
-#endif
-#ifdef __linux__
-#ifndef __GLIBC__
-#define MAP_ANON MAP_ANONYMOUS
-extern caddr_t mmap __P ((caddr_t __addr, size_t __len,
-			  int __prot, int __flags, int __fd, off_t __off));
-extern int munmap __P ((caddr_t __addr, size_t __len));
-#include <linux/mman.h>
-#else
-#include <sys/mman.h>
-#endif
-#endif
 #include <signal.h>
 #include <sys/stat.h>
 #ifdef __linux__
-#if __GLIBC__ > 1
+#if GLIBC_VERSION_CODE >= 2000
 #include <sys/vt.h>
-#include <sys/kd.h>
 #else
-#include <linux/vt.h>
-#include <linux/kd.h>
+#include "Linux/vt.h"
 #endif
-#endif
-#ifdef __NetBSD__
-#include <machine/pcvt_ioctl.h>
+#include <sys/kd.h>
 #endif
 
 #define INIT_C2TRAP
@@ -86,6 +68,8 @@ extern int munmap __P ((caddr_t __addr, size_t __len));
 #include "trident.h"
 #include "et4000.h"
 #include "priv.h"
+#include "mapping.h"
+#include "timers.h"
 
 extern void child_close_mouse ();
 extern void child_open_mouse ();
@@ -137,19 +121,6 @@ forbid_switch (void)
  */
 void check_console(void) {
 
-#ifdef __NetBSD__
-    struct screeninfo info;
-    int result;
-
-    config.console=0;
-    console_fd = STDIN_FILENO;     /* not STDOUT_FILENO */
-
-    if (ioctl(console_fd, VGAGETSCREEN, &info)==0) {
-	/* it's zero-based, we use 1-based */
-	scr_state.console_no = info.screen_no + 1;
-        config.console=1;
-    }
-#endif
 
 #ifdef __linux__
     struct stat chkbuf;
@@ -229,6 +200,7 @@ SIGACQUIRE_call (void)
       get_video_ram (WAIT);
       set_dos_video ();
       /*      if (config.vga) dos_unpause(); */
+      unfreeze_dosemu();
     }
   parent_open_mouse ();
 }
@@ -247,10 +219,6 @@ acquire_vt (int sig, struct sigcontext_struct context)
 
   v_printf ("VID: Acquiring VC\n");
   forbid_switch ();
-#ifdef __NetBSD__
-  if (config.console_keyb)
-      set_raw_mode();
-#endif
   if (ioctl (console_fd, VT_RELDISP, VT_ACKACQ))	/* switch acknowledged */
     v_printf ("VT_RELDISP failed (or was queued)!\n");
   allow_switch ();
@@ -346,11 +314,8 @@ SIGRELEASE_call (void)
 
 	  /*      if (config.vga) dos_pause(); */
 	  scr_state.current = 0;
+	  freeze_dosemu();
 	}
-#ifdef __NetBSD__
-      if (config.console_keyb)
-	  clear_raw_mode();
-#endif
     }
 
   scr_state.current = 0;	/* our console is no longer current */
@@ -442,12 +407,9 @@ get_video_ram (int waitflag)
 	{
 	  vgabuf = malloc (GRAPH_SIZE);
 	  memcpy (vgabuf, (caddr_t) GRAPH_BASE, GRAPH_SIZE);
-	  graph_mem = (char *) mmap ((caddr_t) GRAPH_BASE,
-				     GRAPH_SIZE,
-				     PROT_EXEC | PROT_READ | PROT_WRITE,
-				     MAP_PRIVATE | MAP_FIXED | MAP_ANON,
-				     -1,
-				     0);
+	  graph_mem = (char *)mmap_mapping(MAPPING_VC | MAPPING_SCRATCH,
+			(caddr_t) GRAPH_BASE, GRAPH_SIZE,
+			PROT_EXEC | PROT_READ | PROT_WRITE, 0);
 	  memcpy ((caddr_t) GRAPH_BASE, vgabuf, GRAPH_SIZE);
 	}
     }
@@ -458,12 +420,9 @@ get_video_ram (int waitflag)
 
       if (scr_state.mapped)
 	{
-	  graph_mem = (char *) mmap ((caddr_t) scr_state.virt_address,
-				     TEXT_SIZE,
-				     PROT_EXEC | PROT_READ | PROT_WRITE,
-				     MAP_PRIVATE | MAP_FIXED | MAP_ANON,
-				     -1,
-				     0);
+	  graph_mem = (char *)mmap_mapping(MAPPING_VC | MAPPING_SCRATCH,
+			(caddr_t) scr_state.virt_address, TEXT_SIZE,
+			PROT_EXEC | PROT_READ | PROT_WRITE, 0);
 	  memcpy (scr_state.virt_address, textbuf, TEXT_SIZE);
 	}
     }
@@ -478,15 +437,10 @@ get_video_ram (int waitflag)
 	  /*      else error("ERROR: no dosemu_regs.mem!\n"); */
 	}
       g_printf ("mapping GRAPH_BASE\n");
-      open_kmem ();
-      graph_mem = (char *) mmap ((caddr_t) GRAPH_BASE,
-				 (size_t) (GRAPH_SIZE),
-				 PROT_READ | PROT_WRITE,
-				 MAP_SHARED | MAP_FIXED,
-				 mem_fd,
-				 GRAPH_BASE);
+      graph_mem = (char *)mmap_mapping(MAPPING_VC | MAPPING_KMEM,
+			(caddr_t) GRAPH_BASE, GRAPH_SIZE,
+			PROT_READ | PROT_WRITE, (caddr_t)GRAPH_BASE);
 
-      close_kmem ();
       /* the code below is done by the video save/restore code */
     }
   else
@@ -497,26 +451,10 @@ get_video_ram (int waitflag)
 
       g_printf ("mapping PAGE_ADDR\n");
 
-      open_kmem ();
-#if 0
-      /* Map CGA, etc text memory to HGA memory.
-       Useful for debugging systems with HGA or MDA cards.
-     */
-      graph_mem = (char *) mmap ((caddr_t) 0xb8000,
-				 TEXT_SIZE,
-				 PROT_READ | PROT_WRITE,
-				 MAP_SHARED | MAP_FIXED,
-				 mem_fd,
-				 phys_text_base);
-#endif	/* 0 */
-
-      graph_mem = (char *) mmap ((caddr_t) PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)),
-				 TEXT_SIZE,
-				 PROT_READ | PROT_WRITE,
-				 MAP_SHARED | MAP_FIXED,
-				 mem_fd,
-				 phys_text_base);
-      close_kmem ();
+      graph_mem = (char *)mmap_mapping(MAPPING_VC | MAPPING_KMEM,
+			(caddr_t)PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)),
+			TEXT_SIZE, PROT_READ | PROT_WRITE,
+			(caddr_t)phys_text_base);
 
       if ((long) graph_mem < 0)
 	{
@@ -556,14 +494,9 @@ put_video_ram (void)
 
       if (config.vga)
 	{
-	  open_kmem();
-	  graph_mem = (char *) mmap ((caddr_t) GRAPH_BASE,
-				     GRAPH_SIZE,
-				     PROT_EXEC | PROT_READ | PROT_WRITE,
-				     MAP_PRIVATE | MAP_FIXED | MAP_ANON,
-				     -1,
-				     0);
-	  close_kmem();
+	  graph_mem = (char *)mmap_mapping(MAPPING_VC | MAPPING_SCRATCH,
+				(caddr_t) GRAPH_BASE, GRAPH_SIZE,
+				PROT_EXEC | PROT_READ | PROT_WRITE, 0);
 	  if (dosemu_regs.mem && READ_BYTE(BIOS_VIDEO_MODE) == 3 && READ_BYTE(BIOS_CURRENT_SCREEN_PAGE) < 8) {
 	    memcpy ((caddr_t) PAGE_ADDR(0), dosemu_regs.mem, dosemu_regs.save_mem_size[0]);
 	  }
@@ -571,12 +504,10 @@ put_video_ram (void)
       else
 	{
 	  memcpy (putbuf, scr_state.virt_address, TEXT_SIZE);
-	  graph_mem = (char *) mmap ((caddr_t) scr_state.virt_address,
-				     TEXT_SIZE,
-				     PROT_EXEC | PROT_READ | PROT_WRITE,
-				     MAP_PRIVATE | MAP_FIXED | MAP_ANON,
-				     -1,
-				     0);
+	  graph_mem = (char *)mmap_mapping(MAPPING_VC | MAPPING_SCRATCH,
+				(caddr_t)scr_state.virt_address,
+				TEXT_SIZE, PROT_EXEC | PROT_READ | PROT_WRITE,
+				0);
 	  memcpy (scr_state.virt_address, putbuf, TEXT_SIZE);
 	}
 
@@ -642,11 +573,6 @@ open_kmem (void)
      * and /dev/mem is the identity-mapped (i.e. physical addressed)
      * memory. Currently under Linux, both are the same.
      */
-    /* Under NetBSD, /dev/mem is the physical memory space.  /dev/kmem
-     * is the kernel's mapping to the physical space.  Therefore, we want
-     * /dev/mem.  Since this won't cause problems on linux, I've changed
-     * everything to use /dev/mem.
-     */
 
   kmem_open_count++;
 
@@ -692,14 +618,6 @@ vc_active (void)
   return ((vtstat.v_active == scr_state.console_no));
 #endif
 
-#ifdef __NetBSD__
-  int active;
-
-  g_printf ("VC_ACTIVE!\n");
-  ioctl (console_fd, VT_GETACTIVE, &active);
-  g_printf ("VC_ACTIVE: ours: %d, active: %d\n", scr_state.console_no, active);
-  return ((active == scr_state.console_no));
-#endif
 }
 
 void
@@ -866,6 +784,7 @@ set_regs (u_char regs[])
 {
   int i;
 
+  emu_video_retrace_off();
   /* port_out(dosemu_regs.regs[FCR], FCR_W); */
   /* update misc output register */
   port_out (regs[MIS], MIS_W);
@@ -917,6 +836,7 @@ port_out(dosemu_regs.regs[SEQI], SEQ_I); */
   /* v_printf("CRTI=0x%02x\n",dosemu_regs.regs[CRTI]);
 v_printf("GRAI=0x%02x\n",dosemu_regs.regs[GRAI]);
 v_printf("SEQI=0x%02x\n",dosemu_regs.regs[SEQI]); */
+  emu_video_retrace_on();
   return (0);
 }
 
@@ -925,7 +845,9 @@ v_printf("SEQI=0x%02x\n",dosemu_regs.regs[SEQI]); */
 static inline void
 reset_att (void)
 {
+  emu_video_retrace_off();
   port_in (IS1_R);
+  emu_video_retrace_on();
 }
 
 
@@ -946,7 +868,7 @@ video_port_in (ioport_t port)
       return (dosemu_regs.regs[CRTI]);
     case CRT_DC:
     case CRT_DM:
-      if (dosemu_regs.regs[CRTI] <= 0x18)
+      if (dosemu_regs.regs[CRTI] < CRT_C)
 	{
 	  v_printf ("Read Data at CRTC Index 0x%02x = 0x%02x \n", dosemu_regs.regs[CRTI], dosemu_regs.regs[CRT + dosemu_regs.regs[CRTI]]);
 	  return (dosemu_regs.regs[CRT + dosemu_regs.regs[CRTI]]);
@@ -958,7 +880,7 @@ video_port_in (ioport_t port)
       v_printf ("Read Index GRAI 0x%02x\n", GRAI);
       return (dosemu_regs.regs[GRAI]);
     case GRA_D:
-      if (dosemu_regs.regs[GRAI] < 0x0a)
+      if (dosemu_regs.regs[GRAI] < GRA_C)
 	{
 	  v_printf ("Read Data at GRA  Index 0x%02x = 0x%02x \n", dosemu_regs.regs[GRAI], dosemu_regs.regs[GRA + dosemu_regs.regs[GRAI]]);
 	  return (dosemu_regs.regs[GRA + dosemu_regs.regs[GRAI]]);
@@ -970,7 +892,7 @@ video_port_in (ioport_t port)
       v_printf ("Read Index SEQI 0x%02x\n", SEQI);
       return (dosemu_regs.regs[SEQI]);
     case SEQ_D:
-      if (dosemu_regs.regs[SEQI] < 0x07)
+      if (dosemu_regs.regs[SEQI] < SEQ_C)
 	{
 	  v_printf ("Read Data at SEQ Index 0x%02x = 0x%02x \n", dosemu_regs.regs[SEQI], dosemu_regs.regs[SEQ + dosemu_regs.regs[SEQI]]);
 	  return (dosemu_regs.regs[SEQ + dosemu_regs.regs[SEQI]]);
@@ -1031,7 +953,7 @@ video_port_out (ioport_t port, u_char value)
       break;
     case CRT_DC:
     case CRT_DM:
-      if (dosemu_regs.regs[CRTI] <= 0x18)
+      if (dosemu_regs.regs[CRTI] < CRT_C)
 	{
 	  v_printf ("Write Data at CRTC Index 0x%02x = 0x%02x \n", dosemu_regs.regs[CRTI], value);
 	  dosemu_regs.regs[CRT + dosemu_regs.regs[CRTI]] = value;
@@ -1044,7 +966,7 @@ video_port_out (ioport_t port, u_char value)
       dosemu_regs.regs[GRAI] = value;
       break;
     case GRA_D:
-      if (dosemu_regs.regs[GRAI] < 0x0a)
+      if (dosemu_regs.regs[GRAI] < GRA_C)
 	{
 	  v_printf ("Write Data at GRAI Index 0x%02x = 0x%02x \n", dosemu_regs.regs[GRAI], value);
 	  dosemu_regs.regs[GRA + dosemu_regs.regs[GRAI]] = value;
@@ -1057,7 +979,7 @@ video_port_out (ioport_t port, u_char value)
       dosemu_regs.regs[SEQI] = value;
       break;
     case SEQ_D:
-      if (dosemu_regs.regs[SEQI] <= 0x05)
+      if (dosemu_regs.regs[SEQI] < SEQ_C)
 	{
 	  v_printf ("Write Data at SEQ Index 0x%02x = 0x%02x \n", dosemu_regs.regs[SEQI], value);
 	  dosemu_regs.regs[SEQ + dosemu_regs.regs[SEQI]] = value;
@@ -1260,7 +1182,7 @@ install_int_10_handler (void)
     }
 #else
   if (config.vbios_seg == 0xe000) {
-    extern void bios_f000_int10ptr(), bios_f000();
+    extern void bios_f000_int10ptr();
     ptr = (u_char *)((BIOSSEG << 4) + ((long)bios_f000_int10ptr - (long)bios_f000));
     *((long *)ptr) = 0xe0000003;
     v_printf("VID: new int10 handler at %p\n",ptr);
