@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 
 #ifndef _EMU_KEYB_CLNT_H
 #define _EMU_KEYB_CLNT_H
@@ -5,13 +11,12 @@
 #include "emu.h"
 #include "keyboard.h"
 
-#ifndef NEW_KBD_CODE
-#error keyb_clients.h is for new keyboard code only!
-#endif
 
 /* ASCII 0x20..0x7E */
 extern const Bit8u ascii_keys[];
+extern const u_char latin_to_dos[];
 extern const u_char latin1_to_dos[];
+extern const u_char latin2_to_dos[];
 
 int keyb_client_init();
 void keyb_client_close();
