@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2000 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -32,8 +32,7 @@
 #include <stdio.h>
 #include <errno.h>
 #ifdef __linux__
-#include <asm/types.h>
-#include <linux/genhd.h>
+  #include "Linux/genhd.h"
 #endif
 
 #include "config.h"
